# Copyright (c) Meta Platforms, Inc. and affiliates.
# Copyright (c) 2023, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Functions for embedding the drgn CLI."""

import argparse
import builtins
import contextlib
import importlib
import logging
import os
import os.path
from pathlib import Path
import pkgutil
import runpy
import shutil
import sys
from typing import IO, Any, Callable, Dict, Iterator, Optional, Tuple, Type, TypeVar

import drgn
from drgn.internal.repl import interact, readline
from drgn.internal.rlcompleter import Completer
from drgn.internal.sudohelper import open_via_sudo

__all__ = ("default_globals", "display_str", "run_interactive", "version_header")

logger = logging.getLogger("drgn")

# The list of attributes from the drgn module which are imported and inserted
# into the global namespace for interactive debugging.
_DRGN_GLOBALS = [
    "FaultError",
    "NULL",
    "Object",
    "alignof",
    "cast",
    "container_of",
    "execscript",
    "implicit_convert",
    "offsetof",
    "reinterpret",
    "sizeof",
    "search_memory",
    "search_memory_regex",
    "search_memory_u16",
    "search_memory_u32",
    "search_memory_u64",
    "search_memory_word",
    "source_location",
    "stack_trace",
]


def _is_tty(file: IO[Any]) -> bool:
    try:
        return os.isatty(file.fileno())
    except (AttributeError, OSError):
        return False


class _LogFormatter(logging.Formatter):
    _LEVELS = (
        (logging.DEBUG, "debug", "\033[36m", "\033[m", ""),
        (logging.INFO, "info", "\033[32m", "\033[m", ""),
        (logging.WARNING, "warning", "\033[33m", "\033[m", ""),
        (logging.ERROR, "error", "\033[31m", "\033[m", ""),
        (logging.CRITICAL, "critical", "\033[31;1m", "\033[0;1m", "\033[m"),
    )

    def __init__(self, color: bool) -> None:
        if color:
            levels = {
                level: (f"{level_prefix}{level_name}:{message_prefix}", message_suffix)
                for level, level_name, level_prefix, message_prefix, message_suffix in self._LEVELS
            }
        else:
            levels = {
                level: (f"{level_name}:", "")
                for level, level_name, _, _, _ in self._LEVELS
            }
        default_prefix = "%(levelname)s:"

        self._drgn_formatters = {
            level: logging.Formatter(f"{prefix} %(message)s{suffix}")
            for level, (prefix, suffix) in levels.items()
        }
        self._default_drgn_formatter = logging.Formatter(
            f"{default_prefix} %(message)s"
        )

        self._other_formatters = {
            level: logging.Formatter(f"{prefix}%(name)s: %(message)s{suffix}")
            for level, (prefix, suffix) in levels.items()
        }
        self._default_other_formatter = logging.Formatter(
            f"{default_prefix}%(name)s: %(message)s"
        )

    def format(self, record: logging.LogRecord) -> str:
        if record.name == "drgn":
            formatter = self._drgn_formatters.get(
                record.levelno, self._default_drgn_formatter
            )
        else:
            formatter = self._other_formatters.get(
                record.levelno, self._default_other_formatter
            )
        return formatter.format(record)


def _set_log_handler() -> None:
    handler = logging.StreamHandler()
    handler.setFormatter(_LogFormatter(_is_tty(sys.stderr)))
    logging.getLogger().addHandler(handler)


def version_header() -> str:
    """
    Return the version header printed at the beginning of a drgn session.

    The :func:`run_interactive()` function does not include this banner at the
    beginning of an interactive session. Use this function to retrieve one line
    of text to add to the beginning of the drgn banner, or print it before
    calling :func:`run_interactive()`.
    """
    python_version = ".".join(str(v) for v in sys.version_info[:3])
    debuginfod = f'with{"" if drgn._have_debuginfod else "out"} debuginfod'
    if drgn._enable_dlopen_debuginfod:
        debuginfod += " (dlopen)"
    json_c = f'with{"" if drgn._with_json_c else "out"} json-c'
    libkdumpfile = f'with{"" if drgn._with_libkdumpfile else "out"} libkdumpfile'
    lzma = f'with{"" if drgn._with_lzma else "out"} lzma'
    pcre2 = f'with{"" if drgn._with_pcre2 else "out"} pcre2'
    return f"drgn {drgn.__version__} (using Python {python_version}, elfutils {drgn._elfutils_version}, {debuginfod}, {json_c}, {libkdumpfile}, {lzma}, {pcre2})"


def default_globals(prog: drgn.Program) -> Dict[str, Any]:
    """
    Return the default globals for an interactive drgn session

    :param prog: the program which will be debugged
    :return: a dict of globals
    """
    # Don't forget to update the default banner in run_interactive()
    # with any new additions.
    init_globals: Dict[str, Any] = {
        "prog": prog,
        "drgn": drgn,
        "__name__": "__main__",
        "__doc__": None,
    }
    for attr in _DRGN_GLOBALS:
        init_globals[attr] = getattr(drgn, attr)
    module = importlib.import_module("drgn.helpers.common")
    for name in module.__dict__["__all__"]:
        init_globals[name] = getattr(module, name)
    if prog.flags & drgn.ProgramFlags.IS_LINUX_KERNEL:
        module = importlib.import_module("drgn.helpers.linux")
        for name in module.__dict__["__all__"]:
            init_globals[name] = getattr(module, name)
    return init_globals


def _set_kernel_with_sudo_fallback(prog: drgn.Program) -> None:
    try:
        prog.set_kernel()
        return
    except PermissionError as e:
        if shutil.which("sudo") is None:
            sys.exit(
                f"{e}\ndrgn debugs the live kernel by default, which requires root"
            )
    prog.set_core_dump(open_via_sudo("/proc/kcore", os.O_RDONLY))


def _identify_script(path: str) -> str:
    EI_NIDENT = 16
    SIZEOF_E_TYPE = 2
    KDUMP_SIGNATURE = b"KDUMP   "
    FLATTENED_SIGNATURE = b"makedumpfile\0\0\0\0"

    with open(path, "rb") as f:
        header = f.read(
            max(
                EI_NIDENT + SIZEOF_E_TYPE,
                len(KDUMP_SIGNATURE),
                len(FLATTENED_SIGNATURE),
            )
        )

    if header.startswith(KDUMP_SIGNATURE) or header.startswith(FLATTENED_SIGNATURE):
        return "core"

    ELFMAG = b"\177ELF"
    EI_DATA = 5
    ELFDATA2LSB = 1
    ELFDATA2MSB = 2
    ET_CORE = 4

    if len(header) < EI_NIDENT + SIZEOF_E_TYPE or header[:4] != ELFMAG:
        return "other"

    if header[EI_DATA] == ELFDATA2LSB:
        byteorder = "little"
    elif header[EI_DATA] == ELFDATA2MSB:
        byteorder = "big"
    else:
        return "elf"

    e_type = int.from_bytes(
        header[EI_NIDENT : EI_NIDENT + SIZEOF_E_TYPE],
        byteorder,  # type: ignore[arg-type]  # python/mypy#9057
    )
    return "core" if e_type == ET_CORE else "elf"


_DISPLAY_STR_TYPES: Tuple[Type[Any], ...] = (
    drgn.SourceLocation,
    drgn.SourceLocationList,
    drgn.StackFrame,
    drgn.StackTrace,
    drgn.Type,
)

T = TypeVar("T")


def display_str(tp: Type[T]) -> Type[T]:
    """
    Decorator for types which should be displayed in the cli using
    :py:class:`str()<str>`

    Many types, like :py:class:`drgn.StackTrace`, are formatted in the CLI using
    :py:class:`str()<str>` rather than :py:func:`repr()`. This activates their
    :py:meth:`__str__()<object.__str__>` method rather than the more verbose
    :py:meth:`__repr__()<object.__repr__>`, which results in a more
    user-friendly output. To achieve this behavior with your own types, apply
    this decorator::

        >>> @drgn.cli.display_str
        ... class Task(NamedTuple):
        ...     pid: int
        ...     comm: str
        ...     def __str__(self):
        ...         return f"PID: {self.pid:6d}  COMM: {self.comm}"
        ...
        >>> print(repr(Task(1, "init")))
        Task(pid=1, comm='init')
        >>> print(str(Task(1, "init")))
        PID:      1  COMM: init
        >>> Task(1, "init")
        PID:      1  COMM: init


    :param tp: type to register for display with :py:class:`str`
    :returns: the same type
    """
    global _DISPLAY_STR_TYPES
    _DISPLAY_STR_TYPES = _DISPLAY_STR_TYPES + (tp,)
    return tp


def _displayhook(value: Any) -> None:
    if value is None:
        return
    setattr(builtins, "_", None)
    if isinstance(value, drgn.Object):
        try:
            text = value.format_(columns=shutil.get_terminal_size((0, 0)).columns)
        except drgn.FaultError as e:
            logger.warning("can't print value: %s", e)
            text = repr(value)
    elif isinstance(value, _DISPLAY_STR_TYPES):
        text = str(value)
    else:
        text = repr(value)
    try:
        sys.stdout.write(text)
    except UnicodeEncodeError:
        encoded = text.encode(sys.stdout.encoding, "backslashreplace")
        if hasattr(sys.stdout, "buffer"):
            sys.stdout.buffer.write(encoded)
        else:
            text = encoded.decode(sys.stdout.encoding, "strict")
            sys.stdout.write(text)
    sys.stdout.write("\n")
    setattr(builtins, "_", value)


def _bool_options(value: bool) -> Dict[str, Tuple[str, bool]]:
    return {
        option: ("try_" + option.replace("-", "_"), value)
        for option in (
            "module-name",
            "build-id",
            "debug-link",
            "procfs",
            "embedded-vdso",
            "reuse",
            "supplementary",
        )
    }


class _TrySymbolsByBaseAction(argparse.Action):
    _enable: bool
    _finder = ("disable_debug_info_finders", "enable_debug_info_finders")

    _options = (
        {
            **_bool_options(False),
            "kmod": ("try_kmod", drgn.KmodSearchMethod.NONE),
        },
        {
            **_bool_options(True),
            "kmod=depmod": ("try_kmod", drgn.KmodSearchMethod.DEPMOD),
            "kmod=walk": ("try_kmod", drgn.KmodSearchMethod.WALK),
            "kmod=depmod-or-walk": ("try_kmod", drgn.KmodSearchMethod.DEPMOD_OR_WALK),
            "kmod=depmod-and-walk": ("try_kmod", drgn.KmodSearchMethod.DEPMOD_AND_WALK),
        },
    )

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs["dest"] = argparse.SUPPRESS
        super().__init__(*args, **kwargs)

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: Optional[str] = None,
    ) -> None:
        for value in values.split(","):
            try:
                option_name, option_value = self._options[self._enable][value]
            except KeyError:
                # Raise an error if passed an option meant for the opposite
                # argument.
                if value in self._options[not self._enable]:
                    raise argparse.ArgumentError(self, f"invalid option: {value!r}")

                if not hasattr(namespace, self._finder[self._enable]):
                    setattr(namespace, self._finder[self._enable], {})
                getattr(namespace, self._finder[self._enable])[value] = None

                if hasattr(namespace, self._finder[not self._enable]):
                    getattr(namespace, self._finder[not self._enable]).pop(value, None)
            else:
                if not hasattr(namespace, "debug_info_options"):
                    namespace.debug_info_options = {}
                namespace.debug_info_options[option_name] = option_value


class _TrySymbolsByAction(_TrySymbolsByBaseAction):
    _enable = True


class _NoSymbolsByAction(_TrySymbolsByBaseAction):
    _enable = False


def _load_debugging_symbols(prog: drgn.Program, args: argparse.Namespace) -> None:
    enable_debug_info_finders = getattr(args, "enable_debug_info_finders", ())
    disable_debug_info_finders = getattr(args, "disable_debug_info_finders", ())
    if enable_debug_info_finders or disable_debug_info_finders:
        debug_info_finders = prog.enabled_debug_info_finders()
        registered_debug_info_finders = prog.registered_debug_info_finders()

        unknown_finders = []

        for finder in enable_debug_info_finders:
            if finder not in debug_info_finders:
                if finder in registered_debug_info_finders:
                    debug_info_finders.append(finder)
                else:
                    unknown_finders.append(finder)

        for finder in disable_debug_info_finders:
            try:
                debug_info_finders.remove(finder)
            except ValueError:
                if finder not in registered_debug_info_finders:
                    unknown_finders.append(finder)

        if unknown_finders:
            if len(unknown_finders) == 1:
                unknown_finders_repr = repr(unknown_finders[0])
            elif len(unknown_finders) == 2:
                unknown_finders_repr = (
                    f"{unknown_finders[0]!r} or {unknown_finders[1]!r}"
                )
            elif len(unknown_finders) > 2:
                unknown_finders = [repr(finder) for finder in unknown_finders]
                unknown_finders[-1] = "or " + unknown_finders[-1]
                unknown_finders_repr = ", ".join(unknown_finders)
            logger.warning(
                "no matching debugging information finders or options for %s",
                unknown_finders_repr,
            )

        prog.set_enabled_debug_info_finders(debug_info_finders)

    debug_info_options = getattr(args, "debug_info_options", None)
    if debug_info_options:
        for option, value in debug_info_options.items():
            setattr(prog.debug_info_options, option, value)

    def directories_option(arg_name: str, option_name: Optional[str] = None) -> None:
        if option_name is None:
            option_name = arg_name
        arg = getattr(args, arg_name)
        no_default = getattr(args, "no_default_" + arg_name)
        if arg is not None:
            if no_default:
                setattr(prog.debug_info_options, option_name, arg)
            else:
                setattr(
                    prog.debug_info_options,
                    option_name,
                    tuple(arg) + getattr(prog.debug_info_options, option_name),
                )
        elif no_default:
            setattr(prog.debug_info_options, option_name, ())

    directories_option("debug_directories", "directories")
    directories_option("debug_link_directories")
    directories_option("kernel_directories")

    if args.default_symbols is None:
        args.default_symbols = {"default": True, "main": True}
    try:
        prog.load_debug_info(args.symbols, **args.default_symbols)
    except drgn.MissingDebugInfoError as e:
        if args.default_symbols.get("main"):
            try:
                main_module = prog.main_module()
                critical = (
                    main_module.wants_debug_file() or main_module.wants_loaded_file()
                )
            except LookupError:
                critical = True
        else:
            critical = False
        logger.log(logging.CRITICAL if critical else logging.WARNING, "%s", e)

    if args.extra_symbols:
        for extra_symbol_path in args.extra_symbols:
            extra_symbol_path = os.path.abspath(extra_symbol_path)
            prog.extra_module(extra_symbol_path, create=True).try_file(
                extra_symbol_path
            )


def _main() -> None:
    _set_log_handler()

    version = version_header()
    parser = argparse.ArgumentParser(prog="drgn", description="Programmable debugger")

    program_group = parser.add_argument_group(
        title="program selection",
    ).add_mutually_exclusive_group()
    program_group.add_argument(
        "-k", "--kernel", action="store_true", help="debug the running kernel (default)"
    )
    program_group.add_argument(
        "-c", "--core", metavar="PATH", type=str, help="debug the given core dump"
    )
    program_group.add_argument(
        "-p",
        "--pid",
        metavar="PID",
        type=int,
        help="debug the running process with the given PID",
    )
    program_group.add_argument(
        "--qemu",
        metavar="ADDRESS",
        type=str,
        help="debug a QEMU guest over the QEMU Machine Protocol (QMP) "
        "at the given address, which may be a Unix domain socket path "
        "or a TCP address as host:port",
    )

    symbol_group = parser.add_argument_group("debugging symbols")
    symbol_group.add_argument(
        "-s",
        "--symbols",
        metavar="PATH",
        type=str,
        action="append",
        help="load debugging symbols from the given file. "
        "If the file does not correspond to a loaded executable, library, or module, "
        "then it is ignored. This option may be given more than once",
    )
    default_symbols_group = symbol_group.add_mutually_exclusive_group()
    default_symbols_group.add_argument(
        "--main-symbols",
        dest="default_symbols",
        action="store_const",
        const={"main": True},
        help="only load debugging symbols for the main executable "
        "and those added with -s or --extra-symbols",
    )
    default_symbols_group.add_argument(
        "--no-default-symbols",
        dest="default_symbols",
        action="store_const",
        const={},
        help="don't load any debugging symbols that were not explicitly added "
        "with -s or --extra-symbols",
    )
    symbol_group.add_argument(
        "--extra-symbols",
        metavar="PATH",
        type=str,
        action="append",
        help="load additional debugging symbols from the given file, "
        "which is assumed not to correspond to a loaded executable, library, or module. "
        "This option may be given more than once",
    )
    symbol_group.add_argument(
        "--try-symbols-by",
        metavar="METHOD[,METHOD...]",
        action=_TrySymbolsByAction,
        help="enable loading debugging symbols using the given methods. "
        "Choices are debugging information finder names "
        "(standard, debuginfod, or any added by plugins) "
        "or debugging information options ("
        + ", ".join(_TrySymbolsByBaseAction._options[True])
        + "). "
        "This option may be given more than once",
    )
    symbol_group.add_argument(
        "--no-symbols-by",
        metavar="METHOD[,METHOD...]",
        action=_NoSymbolsByAction,
        help="disable loading debugging symbols using the given methods. "
        "Choices are debugging information finder names "
        "(standard, debuginfod, or any added by plugins) "
        "or debugging information options ("
        + ", ".join(_TrySymbolsByBaseAction._options[False])
        + "). "
        "This option may be given more than once",
    )

    directories_group = parser.add_argument_group("debugging symbol directories")
    directories_group.add_argument(
        "--debug-directory",
        dest="debug_directories",
        metavar="PATH",
        type=str,
        action="append",
        help="search for debugging symbols in the given directory. "
        "This option may be given more than once",
    )
    directories_group.add_argument(
        "--no-default-debug-directories",
        action="store_true",
        help="don't search for debugging symbols "
        "in the standard directories or those added by plugins",
    )
    directories_group.add_argument(
        "--debug-link-directory",
        dest="debug_link_directories",
        metavar="PATH",
        type=str,
        action="append",
        help="search for debugging symbols by debug link in the given directory. "
        "$ORIGIN is replaced with the directory containing the loaded file. "
        "This option may be given more than once",
    )
    directories_group.add_argument(
        "--no-default-debug-link-directories",
        action="store_true",
        help="don't search for debugging symbols by debug link "
        "in the standard directories or those added by plugins",
    )
    directories_group.add_argument(
        "--kernel-directory",
        dest="kernel_directories",
        metavar="PATH",
        type=str,
        action="append",
        help="search for the kernel image and loadable kernel modules in the given directory. "
        "This option may be given more than once",
    )
    directories_group.add_argument(
        "--no-default-kernel-directories",
        action="store_true",
        help="don't search for the kernel image and loadable kernel modules "
        "in the standard directories or those added by plugins",
    )

    advanced_group = parser.add_argument_group("advanced")
    advanced_group.add_argument(
        "--architecture",
        metavar="ARCH",
        choices=[a.name for a in drgn.Architecture]
        + [a.name.lower() for a in drgn.Architecture],
        help="set the program architecture, in case it can't be auto-detected",
    )
    advanced_group.add_argument(
        "--vmcoreinfo",
        type=str,
        metavar="PATH",
        help="path to vmcoreinfo file (overrides any already present in the file)",
    )
    parser.add_argument(
        "--log-level",
        choices=["debug", "info", "warning", "error", "critical", "none"],
        default="warning",
        help="log messages of at least the given level to standard error (default: warning)",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="log_level",
        action="store_const",
        const="none",
        help="don't print any logs or download progress",
    )
    parser.add_argument(
        "-e",
        dest="exec",
        metavar="CODE",
        help="an expression or statement to evaluate, instead of running in interactive mode",
    )
    parser.add_argument(
        "args",
        metavar="ARG",
        type=str,
        nargs=argparse.REMAINDER,
        help="script to execute instead of running in interactive mode "
        "(unless -e is given) and arguments to pass",
    )
    parser.add_argument("--version", action="version", version=version)

    args = parser.parse_args()

    script = bool(args.exec is None and args.args)
    interactive = bool(args.exec is None and not args.args and _is_tty(sys.stdin))
    if script:
        # A common mistake users make is running drgn $core_dump, which tries
        # to run $core_dump as a Python script. Rather than failing later with
        # some inscrutable syntax or encoding error, try to catch this early
        # and provide a helpful message.
        try:
            script_type = _identify_script(args.args[0])
        except OSError as e:
            sys.exit(str(e))
        if script_type == "core":
            sys.exit(
                f"error: {args.args[0]} is a core dump\n"
                f'Did you mean "-c {args.args[0]}"?'
            )
        elif script_type == "elf":
            sys.exit(f"error: {args.args[0]} is a binary, not a drgn script")
    elif interactive:
        print(version, file=sys.stderr, flush=True)

    if args.log_level == "none":
        logger.setLevel(logging.CRITICAL + 1)
    else:
        logger.setLevel(args.log_level.upper())

    platform = None
    if args.architecture:
        platform = drgn.Platform(drgn.Architecture[args.architecture.upper()])

    vmcoreinfo = None
    if args.vmcoreinfo is not None:
        with open(args.vmcoreinfo, "rb") as f:
            vmcoreinfo = f.read()

    prog = drgn.Program(platform=platform, vmcoreinfo=vmcoreinfo)
    try:
        if args.core is not None:
            prog.set_core_dump(args.core)
        elif args.pid is not None:
            try:
                prog.set_pid(args.pid or os.getpid())
            except PermissionError as e:
                sys.exit(
                    f"{e}\nerror: attaching to live process requires ptrace attach permissions"
                )
        elif args.qemu is not None:
            prog.set_qemu_qmp(args.qemu)
        else:
            _set_kernel_with_sudo_fallback(prog)
    except OSError as e:
        sys.exit(str(e))
    except ValueError as e:
        # E.g., "not an ELF core file"
        sys.exit(f"error: {e}")

    _load_debugging_symbols(prog, args)

    if interactive:
        run_interactive(prog)
    else:
        drgn.set_default_prog(prog)
        if script:
            sys.argv = args.args
            script_path = args.args[0]
            if pkgutil.get_importer(script_path) is None:
                sys.path.insert(0, os.path.dirname(os.path.abspath(script_path)))
            runpy.run_path(
                script_path, init_globals={"prog": prog}, run_name="__main__"
            )
        else:
            sys.path.insert(0, "")
            exec_globals = default_globals(prog)
            if args.exec is None:
                sys.argv = [""]
                exec_globals["__file__"] = "<stdin>"
                exec(compile(sys.stdin.read(), "<stdin>", "exec"), exec_globals)
            else:
                sys.argv = ["-e"] + args.args
                exec(args.exec, exec_globals)


def _history_file() -> str:
    # Historically, our history file was directly in ~. However, the May 2021
    # XDG Base Directory Specification [1] added a dedicated directory for
    # things like history files: $XDG_STATE_HOME, which defaults to
    # ~/.local/state.
    #
    # We don't move existing history files, but we create new ones in the new
    # location.
    #
    # Note that we use pathlib because it raises an error if a home directory
    # can't be resolved instead of failing silently like os.path.expanduser().
    #
    # [1]: https://specifications.freedesktop.org/basedir/latest/
    path = Path("~/.drgn_history").expanduser()
    if path.exists():
        return str(path)
    return _state_file("history")


def _state_file(name: str) -> str:
    xdg_state_home = os.getenv("XDG_STATE_HOME")
    if xdg_state_home is None:
        path = Path("~/.local/state").expanduser()
    else:
        path = Path(xdg_state_home)
    return str(path / "drgn" / name)


def _read_history(history_file: str) -> None:
    try:
        readline.read_history_file(history_file)
    except FileNotFoundError:
        pass
    except OSError as e:
        logger.warning("could not read history: %s", e)


def _write_history(history_file: str) -> None:
    try:
        os.makedirs(os.path.dirname(history_file), exist_ok=True)
        readline.write_history_file(history_file)
    except OSError as e:
        logger.warning("could not write history: %s", e)


current_history_file = None


@contextlib.contextmanager
def _setup_readline(
    history_file: str, completer: Optional[Callable[[str, int], Optional[str]]] = None
) -> Iterator[None]:
    global current_history_file
    old_history_file = current_history_file
    old_history_length = readline.get_history_length()
    old_completer = readline.get_completer()

    try:
        if current_history_file is not None:
            _write_history(current_history_file)
        readline.clear_history()
        _read_history(history_file)
        current_history_file = history_file

        readline.set_history_length(1000)

        if completer is None:
            readline.parse_and_bind("tab: self-insert")
        else:
            readline.parse_and_bind("tab: complete")
        readline.set_completer(completer)

        try:
            yield
        finally:
            _write_history(history_file)
    finally:
        readline.set_completer(old_completer)
        if old_completer is None:
            readline.parse_and_bind("tab: self-insert")
        else:
            readline.parse_and_bind("tab: complete")

        readline.set_history_length(old_history_length)

        readline.clear_history()
        if old_history_file is not None:
            _read_history(old_history_file)
        current_history_file = old_history_file


def run_interactive(
    prog: drgn.Program,
    banner_func: Optional[Callable[[str], str]] = None,
    globals_func: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None,
    quiet: bool = False,
) -> None:
    """
    Run drgn's :ref:`interactive-mode` until the user exits.

    This function allows your application to embed the same REPL that drgn
    provides when it is run on the command line in interactive mode.

    :param prog: Pre-configured program to run against. Available as a global
        named ``prog`` in the CLI.
    :param banner_func: Optional function to modify the printed banner. Called
        with the default banner, and must return a string to use as the new
        banner. The default banner does not include the drgn version, which can
        be retrieved via :func:`version_header()`.
    :param globals_func: Optional function to modify globals provided to the
        session. Called with a dictionary of default globals, and must return a
        dictionary to use instead.
    :param quiet: Ignored. Will be removed in the future.

    .. note::

        This function uses :mod:`readline` and modifies some settings.
        Unfortunately, it is not possible for it to restore all settings. In
        particular, it clears the ``readline`` history and resets the TAB
        keybinding to the default.

        Applications using ``readline`` should save their history and clear any
        custom settings before calling this function. After calling this
        function, applications should restore their history and settings before
        using ``readline``.
    """
    init_globals = default_globals(prog)
    banner = f"""\
For help, type help(drgn).
>>> import drgn
>>> from drgn import {", ".join(_DRGN_GLOBALS)}
>>> from drgn.helpers.common import *"""
    if prog.flags & drgn.ProgramFlags.IS_LINUX_KERNEL:
        banner += "\n>>> from drgn.helpers.linux import *"
    if banner_func:
        banner = banner_func(banner)
    if globals_func:
        init_globals = globals_func(init_globals)

    old_path = list(sys.path)
    old_displayhook = sys.displayhook
    try:
        old_default_prog = drgn.get_default_prog()
    except drgn.NoDefaultProgramError:
        old_default_prog = None
    had_outer_repl = "outer_repl" in prog.config

    with _setup_readline(_history_file(), Completer(init_globals).complete):
        try:
            sys.path.insert(0, "")
            sys.displayhook = _displayhook

            drgn.set_default_prog(prog)

            if not had_outer_repl:
                prog.config["outer_repl"] = "drgn"

            interact(init_globals, banner)
        finally:
            if not had_outer_repl:
                prog.config.pop("outer_repl", None)
            drgn.set_default_prog(old_default_prog)
            sys.displayhook = old_displayhook
            sys.path[:] = old_path



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Bit Operations
--------------

The ``drgn.helpers.linux.bitops`` module provides helpers for common bit
operations in the Linux kernel.
"""

import operator
from typing import Iterator

from drgn import IntegerLike, Object, sizeof

__all__ = (
    "for_each_clear_bit",
    "for_each_set_bit",
    "test_bit",
)


def for_each_set_bit(bitmap: Object, size: IntegerLike) -> Iterator[int]:
    """
    Iterate over all set (one) bits in a bitmap.

    :param bitmap: ``unsigned long *``
    :param size: Size of *bitmap* in bits.
    """
    size = operator.index(size)
    word_bits = 8 * sizeof(bitmap.type_.type)
    for i in range((size + word_bits - 1) // word_bits):
        word = bitmap[i].value_()
        for j in range(min(word_bits, size - word_bits * i)):
            if word & (1 << j):
                yield (word_bits * i) + j


def for_each_clear_bit(bitmap: Object, size: IntegerLike) -> Iterator[int]:
    """
    Iterate over all clear (zero) bits in a bitmap.

    :param bitmap: ``unsigned long *``
    :param size: Size of *bitmap* in bits.
    """
    size = operator.index(size)
    word_bits = 8 * sizeof(bitmap.type_.type)
    for i in range((size + word_bits - 1) // word_bits):
        word = bitmap[i].value_()
        for j in range(min(word_bits, size - word_bits * i)):
            if not (word & (1 << j)):
                yield (word_bits * i) + j


def test_bit(nr: IntegerLike, bitmap: Object) -> bool:
    """
    Return whether a bit in a bitmap is set.

    :param nr: Bit number.
    :param bitmap: ``unsigned long *``
    """
    nr = operator.index(nr)
    word_bits = 8 * sizeof(bitmap.type_.type)
    return ((bitmap[nr // word_bits].value_() >> (nr & (word_bits - 1))) & 1) != 0



# Copyright (c) Stephen Brennan <stephen@brennan.io>
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Helper for opening a file as root and transmitting it via unix socket"""
import array
from pathlib import Path
import pickle
import socket
import subprocess
import sys
import tempfile
from typing import Union

_OPEN_VIA_SUDO_COMMAND = r"""
import array
import os
import pickle
import socket
import sys

sockpath = sys.argv[1]
filename = sys.argv[2]
flags = int(sys.argv[3])
mode = int(sys.argv[4])

sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
sock.connect(sockpath)
try:
    fd = os.open(filename, flags, mode)
    fds = array.array("i", [fd])
    sock.sendmsg(
        [b"success"],
        [(socket.SOL_SOCKET, socket.SCM_RIGHTS, fds)],
    )
except Exception as e:
    sock.sendmsg([pickle.dumps(e)])
"""


def open_via_sudo(
    path: Union[Path, str],
    flags: int,
    mode: int = 0o777,
) -> int:
    """Implements os.open() using sudo to get permissions"""
    # Currently does not support dir_fd argument
    with tempfile.TemporaryDirectory() as td:
        sockpath = Path(td) / "sock"
        with socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM) as sock:
            sock.bind(str(sockpath))
            subprocess.check_call(
                [
                    "sudo",
                    "-p",
                    f"[sudo] password for %p to open {path}: ",
                    sys.executable,
                    "-B",
                    "-c",
                    _OPEN_VIA_SUDO_COMMAND,
                    sockpath,
                    path,
                    str(flags),
                    str(mode),
                ],
            )
            fds = array.array("i")
            msg, ancdata, flags, addr = sock.recvmsg(
                4096, socket.CMSG_SPACE(fds.itemsize)
            )
            for level, typ, data in ancdata:
                if level == socket.SOL_SOCKET and typ == socket.SCM_RIGHTS:
                    data = data[: fds.itemsize]
                    fds.frombytes(data)
                    return fds[0]
            raise pickle.loads(msg)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Functions for defining Linux kernel-specific commands."""

from typing import Any, Dict

from _drgn_util.typingutils import copy_func_signature
from drgn import Program, ProgramFlags
from drgn.commands import command, custom_command, raw_command


def _program_is_linux_kernel(prog: Program) -> bool:
    return bool(prog.flags & ProgramFlags.IS_LINUX_KERNEL)


def _set_enabled_if_linux_kernel(kwargs: Dict[str, Any]) -> None:
    if "enabled" in kwargs:
        old_enabled = kwargs["enabled"]
        kwargs["enabled"] = lambda prog: _program_is_linux_kernel(prog) and old_enabled(
            prog
        )
    else:
        kwargs["enabled"] = _program_is_linux_kernel


@copy_func_signature(command)
def linux_kernel_command(*args: Any, **kwargs: Any) -> Any:
    """
    Like :func:`~drgn.commands.command()`, but only enables the command when
    debugging the Linux kernel.

    If *enabled* is given, it is also checked in addition to the Linux kernel
    check.
    """
    _set_enabled_if_linux_kernel(kwargs)
    return command(*args, **kwargs)


@copy_func_signature(custom_command)
def linux_kernel_custom_command(*args: Any, **kwargs: Any) -> Any:
    """
    Like :func:`linux_kernel_command()` but for
    :func:`~drgn.commands.custom_command()`.
    """
    _set_enabled_if_linux_kernel(kwargs)
    return custom_command(*args, **kwargs)


@copy_func_signature(raw_command)
def linux_kernel_raw_command(*args: Any, **kwargs: Any) -> Any:
    """
    Like :func:`linux_kernel_command()` but for
    :func:`~drgn.commands.raw_command()`.
    """
    _set_enabled_if_linux_kernel(kwargs)
    return raw_command(*args, **kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Built-in commands exported as the "builtin_commands" plugin. All commands built
into drgn should be defined in this package.
"""

import argparse
import importlib
import pkgutil
import subprocess
import types
from typing import Any, Dict, Union

from drgn import Program, execscript
from drgn.commands import (
    CommandArgumentError,
    _parse_py_command,
    _print_py_command_exception,
    argument,
    command,
    custom_command,
    raw_command,
)

# Import all submodules, recursively.
for _module_info in pkgutil.walk_packages(__path__, __name__ + "."):
    importlib.import_module(_module_info.name)


@raw_command(
    description="execute a shell command",
    usage="**sh** [*command*]",
    long_description="""
    If *command* is given, run it with ``sh -c --``. Otherwise, run an
    interactive shell with ``sh -i``.

    In either case, return the command's exit status.
    """,
)
def _cmd_sh(prog: Program, name: str, args: str, **kwargs: Any) -> int:
    if args:
        return subprocess.call(["sh", "-c", "--", args])
    else:
        return subprocess.call(["sh", "-i"])


@custom_command(
    description="execute a python statement and allow shell redirection",
    usage="**py** *code*",
    long_description="""
    Execute the given code, up to the first shell redirection or pipeline
    statement, as Python code.

    For each occurrence of a pipeline operator (``|``) or any redirection
    operator (``<``, ``>``, ``<<``, ``>>``), attempt to parse the preceding text
    as Python code. If the preceding text is syntactically valid code, then
    interpret the remainder of the command as shell redirections or pipelines,
    and execute the Python code with those redirections and pipelines applied.

    The operators above can be used in syntactically valid Python. This means
    you need to be careful when using this function, and ensure that you wrap
    their uses with parentheses.

    For example, consider the command: ``%py field | MY_FLAG | grep foo``. While
    the intent here may be to execute the Python code ``field | MY_FLAG`` and
    pass its result to ``grep``, that is not what will happen. The portion of
    text prior to the first ``|`` is valid Python, so it will be executed, and
    its output piped to the shell pipeline ``MY_FLAG | grep foo``. Instead,
    running ``%py (field | MY_FLAG) | grep foo`` ensures that ``field |
    MY_FLAG`` gets piped to ``grep foo``, because ``(field`` on its own is not
    valid Python syntax.
    """,
    parse=_parse_py_command,
)
def _cmd_py(
    prog: Program,
    name: str,
    code: Union[types.CodeType, SyntaxError, None],
    *,
    globals: Dict[str, Any],
    **kwargs: Any,
) -> None:
    if code is None:
        raise CommandArgumentError("expected code")

    if isinstance(code, SyntaxError):
        _print_py_command_exception(code)
        return

    try:
        exec(code, globals)
    except (Exception, KeyboardInterrupt) as e:
        # Any exception should be formatted just as the interpreter would. This
        # includes keyboard interrupts, but not things like SystemExit or
        # GeneratorExit.
        _print_py_command_exception(e)


@command(
    description="run a drgn script",
    long_description="""
    This loads and runs a drgn script in the current environment. Currently
    defined globals are available to the script, and globals defined by the
    script are added to the environment.
    """,
    arguments=(
        argument("script", help="script file path"),
        argument(
            "args", nargs=argparse.REMAINDER, help="arguments to pass to the script"
        ),
    ),
)
def _cmd_source(
    prog: Program,
    name: str,
    args: argparse.Namespace,
    *,
    globals: Dict[str, Any],
    **kwargs: Any,
) -> None:
    execscript(args.script, *args.args, globals=globals)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
IDR
---

The ``drgn.helpers.linux.idr`` module provides helpers for working with the IDR
data structure in :linux:`include/linux/idr.h`. An IDR provides a mapping from
an ID to a pointer.
"""

import operator
from typing import Iterator, Tuple, Union

from _drgn import _linux_helper_idr_find
from drgn import NULL, IntegerLike, Object, Type, cast, sizeof
from drgn.helpers.linux.radixtree import radix_tree_for_each

__all__ = (
    "idr_find",
    "idr_for_each",
    "idr_for_each_entry",
)

_IDR_BITS = 8
_IDR_MASK = (1 << _IDR_BITS) - 1


def idr_find(idr: Object, id: IntegerLike) -> Object:
    """
    Look up the entry with the given ID in an IDR.

    :param idr: ``struct idr *``
    :param id: Entry ID.
    :return: ``void *`` found entry, or ``NULL`` if not found.
    """
    # Since Linux kernel commit 0a835c4f090a ("Reimplement IDR and IDA using
    # the radix tree") (in v4.11), IDRs are backed by radix trees. Before that,
    # they are a separate data structure. The helper in libdrgn only handles
    # the radix tree version.
    if hasattr(idr, "idr_rt"):
        return _linux_helper_idr_find(idr, id)
    else:
        prog = idr.prog_
        id = operator.index(id)

        if id < 0:
            return NULL(prog, "void *")

        p = idr.top.read_()
        if not p:
            return NULL(prog, "void *")

        n = (p.layer.value_() + 1) * _IDR_BITS
        MAX_IDR_SHIFT = sizeof(prog.type("int")) * 8 - 1
        # Equivalent to id > idr_max(p->layer + 1) in the kernel.
        if id >= (1 << min(n, MAX_IDR_SHIFT)):
            return NULL(prog, "void *")

        while n > 0 and p:
            n -= _IDR_BITS
            p = p.ary[(id >> n) & _IDR_MASK].read_()

        return cast("void *", p)


def idr_for_each(idr: Object) -> Iterator[Tuple[int, Object]]:
    """
    Iterate over all of the pointers in an IDR.

    :param idr: ``struct idr *``
    :return: Iterator of (index, ``void *``) tuples.
    """
    # Since Linux kernel commit 0a835c4f090a ("Reimplement IDR and IDA using
    # the radix tree") (in v4.11), IDRs are backed by radix trees.
    try:
        idr_rt = idr.idr_rt
    except AttributeError:
        voidp_type = idr.prog_.type("void *")

        def aux(p: Object, id: int, n: int) -> Iterator[Tuple[int, Object]]:
            p = p.read_()
            if p:
                if n == 0:
                    yield id, cast(voidp_type, p)
                else:
                    n -= _IDR_BITS
                    for child in p.ary:
                        yield from aux(child, id, n)
                        id += 1 << n

        yield from aux(idr.top, 0, idr.layers.value_() * _IDR_BITS)
    else:
        try:
            base = idr.idr_base.value_()
        except AttributeError:
            base = 0
        for index, entry in radix_tree_for_each(idr_rt.address_of_()):
            yield index + base, entry


def idr_for_each_entry(
    idr: Object, type: Union[str, Type]
) -> Iterator[Tuple[int, Object]]:
    """
    Iterate over all of the entries with the given type in an IDR.

    :param idr: ``struct idr *``
    :param type: Entry type.
    :return: Iterator of (index, ``type *``) tuples.
    """
    prog = idr.prog_
    type = prog.pointer_type(prog.type(type))
    for index, entry in idr_for_each(idr):
        yield index, cast(type, entry)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Boot
----

The ``drgn.helpers.linux.boot`` module provides helpers for inspecting the
Linux kernel boot configuration.
"""

from _drgn import _linux_helper_kaslr_offset, _linux_helper_pgtable_l5_enabled
from drgn import Program
from drgn.helpers.common.prog import takes_program_or_default

__all__ = (
    "kaslr_offset",
    "pgtable_l5_enabled",
)


@takes_program_or_default
def kaslr_offset(prog: Program) -> int:
    """
    Get the kernel address space layout randomization offset (zero if it is
    disabled).
    """
    return _linux_helper_kaslr_offset(prog)


@takes_program_or_default
def pgtable_l5_enabled(prog: Program) -> bool:
    """Return whether 5-level paging is enabled."""
    return _linux_helper_pgtable_l5_enabled(prog)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Kernfs
------

The ``drgn.helpers.linux.kernfs`` module provides helpers for working with the
kernfs pseudo filesystem interface in :linux:`include/linux/kernfs.h`.
"""

import os
from typing import Iterator

from drgn import NULL, Object, Path
from drgn.helpers.linux.rbtree import rbtree_inorder_for_each_entry

__all__ = (
    "kernfs_name",
    "kernfs_parent",
    "kernfs_path",
    "kernfs_root",
    "kernfs_walk",
    "kernfs_children",
)


def kernfs_root(kn: Object) -> Object:
    """
    Get the kernfs root that the given kernfs node belongs to.

    :param kn: ``struct kernfs_node *``
    :return: ``struct kernfs_root *``
    """
    knp = kernfs_parent(kn)
    if knp:
        kn = knp
    return kn.dir.root.read_()


def kernfs_parent(kn: Object) -> Object:
    """
    Get the parent of the given kernfs node.

    :param kn: ``struct kernfs_node *``
    :return: ``struct kernfs_node *``
    """
    # Linux kernel commit 633488947ef6 ("kernfs: Use RCU to access
    # kernfs_node::parent.") (in v6.15) renamed the parent member.
    try:
        return kn.__parent.read_()
    except AttributeError:
        return kn.parent.read_()


def kernfs_name(kn: Object) -> bytes:
    """
    Get the name of the given kernfs node.

    :param kn: ``struct kernfs_node *``
    """
    if not kn:
        return b"(null)"
    return kn.name.string_() if kernfs_parent(kn) else b"/"


def kernfs_path(kn: Object) -> bytes:
    """
    Get full path of the given kernfs node.

    :param kn: ``struct kernfs_node *``
    """
    if not kn:
        return b"(null)"

    root_kn = kernfs_root(kn).kn
    if kn == root_kn:
        return b"/"

    names = []
    while kn != root_kn:
        names.append(kn.name.string_())
        kn = kernfs_parent(kn)
    names.append(root_kn.name.string_())
    names.reverse()

    return b"/".join(names)


def _kernfs_node_type(kn: Object, node_type: str) -> bool:
    KERNFS_TYPE_MASK = 0x000F
    return kn.flags & KERNFS_TYPE_MASK == kn.prog_.constant(node_type)


def _kernfs_follow_symlink(parent: Object) -> Object:
    while _kernfs_node_type(parent, "KERNFS_LINK"):
        parent = parent.symlink.target_kn
    return parent


def kernfs_walk(parent: Object, path: Path, follow_symlinks: bool = True) -> Object:
    """
    Find the kernfs node with the given path from the given parent kernfs node.

    :param parent: ``struct kernfs_node *``
    :param path: Path name.
    :param follow_symlinks: If True (default), all symbolic links encountered
        in the path, including the final component, are followed and the
        function returns the target node. If False, all intermediate symlinks
        are still followed, but if the final component is a symlink, the
        function returns the symlink node itself rather than its target.
    :return: ``struct kernfs_node *`` (``NULL`` if not found)
    """
    kernfs_nodep_type = parent.type_
    kernfs_node_type = kernfs_nodep_type.type
    for name in os.fsencode(path).split(b"/"):
        if not name:
            continue

        parent = _kernfs_follow_symlink(parent)

        for parent in rbtree_inorder_for_each_entry(
            kernfs_node_type, parent.dir.children.address_of_(), "rb"
        ):
            if (
                parent.name.string_() == name
                and not parent.ns  # For now, we don't bother with namespaced kernfs nodes.
            ):
                break
        else:
            return NULL(parent.prog_, kernfs_nodep_type)

    if follow_symlinks:
        parent = _kernfs_follow_symlink(parent)

    return parent


def kernfs_children(kn: Object) -> Iterator[Object]:
    """
    Get an iterator over the children of the given kernfs node.

    :param kn: ``struct kernfs_node *``
    :return: Iterator of ``struct kernfs_node *`` objects.
    :raises ValueError: if the kernfs node is not a directory
    """
    if not _kernfs_node_type(kn, "KERNFS_DIR"):
        raise ValueError("not a directory")

    return rbtree_inorder_for_each_entry(
        "struct kernfs_node", kn.dir.children.address_of_(), "rb"
    )



# Copyright (c) 2026, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Completion Variables
--------------------

The ``drgn.helpers.linux.completion`` module provides helpers for working with
completion variables (``struct completion``) from
:linux:`include/linux/completion.h`.
"""

from typing import Iterator

from drgn import Object
from drgn.helpers.linux.swait import swait_for_each_task
from drgn.helpers.linux.wait import waitqueue_for_each_task

__all__ = (
    "completion_done",
    "completion_for_each_task",
)


def completion_done(completion: Object) -> bool:
    """
    Test if a completion has any waiters.

    :param completion: ``struct completion *``
    """
    return bool(completion.done)


def completion_for_each_task(completion: Object) -> Iterator[Object]:
    """
    Iterate over all tasks waiting on a completion variable.

    :param completion: ``struct completion *``
    :return: Iterator of ``struct task_struct *`` objects.
    """
    wait = completion.wait.address_of_()
    # completion->wait is a simple wait queue since Linux kernel commit
    # a5c6234e1028 ("completion: Use simple wait queues") (in v5.7).
    # Also Linux kernel commit 2055da97389a ("sched/wait: Disambiguate
    # wq_entry->task_list and wq_head->task_list naming") (in v4.13) renamed
    # the task_list member to head.
    # So completion->wait in kernels v5.7 and later and in kernels prior to v4.13,
    # have task_list member, but of different types. So use type of completion->wait
    # to differentiate between completion backends.
    if wait.type_.type_name() == "struct swait_queue_head *":
        return swait_for_each_task(wait)
    else:
        return waitqueue_for_each_task(wait)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Bitmaps
-------

The ``drgn.helpers.linux.bitmap`` module provides helpers for working with
bitmaps from :linux:`include/linux/bitmap.h`.

The following helpers from :mod:`drgn.helpers.linux.bitops` also apply to
bitmaps:

* :func:`~drgn.helpers.linux.bitops.for_each_set_bit()`
* :func:`~drgn.helpers.linux.bitops.for_each_clear_bit()`
* :func:`~drgn.helpers.linux.bitops.test_bit()`
"""

import operator
import sys

from drgn import IntegerLike, Object, sizeof

__all__ = ("bitmap_weight",)

if sys.version_info >= (3, 10):
    _bit_count = int.bit_count  # novermin
else:

    # Fallback for old Python versions. Surprisingly, this is faster than any
    # bit manipulation tricks.
    def _bit_count(n: int) -> int:
        return bin(n).count("1")


def bitmap_weight(bitmap: Object, size: IntegerLike) -> int:
    """
    Return the number of set (one) bits in a bitmap

    :param bitmap: ``unsigned long *``
    :param size: Size of *bitmap* in bits.
    """
    size = operator.index(size)
    word_bits = 8 * sizeof(bitmap.type_.type)

    weight = 0
    for i in range(size // word_bits):
        weight += _bit_count(bitmap[i].value_())

    last_word_bits = size % word_bits
    if last_word_bits:
        weight += _bit_count(
            bitmap[size // word_bits].value_() & ((1 << last_word_bits) - 1)
        )

    return weight



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Memory Management
-----------------

The ``drgn.helpers.linux.mm`` module provides helpers for working with the
Linux memory management (MM) subsystem.

.. _virtual address translation failures:

Helpers that translate virtual addresses or read virtual memory may fail for
multiple reasons:

1. If the address is invalid.
2. If the address is swapped or paged out.
3. If the address is in `high memory
   <https://docs.kernel.org/mm/highmem.html>`_. High memory is only used for
   userspace memory by 32-bit systems with a lot of physical memory, and only
   if ``CONFIG_HIGHMEM`` is enabled.

   3a. If the page table is in high memory. This is only possible if
   ``CONFIG_HIGHPTE`` is enabled.
"""

import operator
import re
from typing import Callable, Iterable, Iterator, List, NamedTuple, Optional, Tuple

from _drgn import (
    _linux_helper_direct_mapping_offset,
    _linux_helper_follow_phys,
    _linux_helper_read_vm,
)
from drgn import (
    NULL,
    Architecture,
    IntegerLike,
    Object,
    ObjectAbsentError,
    ObjectNotFoundError,
    Program,
    TypeKind,
    cast,
    container_of,
    sizeof,
)
from drgn.helpers.common.format import decode_enum_type_flags
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.bitops import for_each_set_bit
from drgn.helpers.linux.device import bus_for_each_dev
from drgn.helpers.linux.fs import d_path
from drgn.helpers.linux.list import list_for_each_entry
from drgn.helpers.linux.mapletree import mt_for_each, mtree_load
from drgn.helpers.linux.mmzone import _highest_present_section_nr, _section_flags
from drgn.helpers.linux.percpu import percpu_counter_sum, percpu_counter_sum_positive
from drgn.helpers.linux.pid import for_each_task_in_group
from drgn.helpers.linux.rbtree import rb_find

__all__ = (
    "PFN_PHYS",
    "PHYS_PFN",
    "PageCompound",
    "PageHead",
    "PageSlab",
    "PageTail",
    "access_process_vm",
    "access_remote_vm",
    "cmdline",
    "compound_head",
    "compound_nr",
    "compound_order",
    "decode_memory_block_state",
    "decode_memory_block_state_value",
    "decode_page_flags",
    "decode_page_flags_value",
    "environ",
    "find_vmap_area",
    "follow_page",
    "follow_pfn",
    "follow_phys",
    "for_each_memory_block",
    "for_each_page",
    "for_each_valid_page_range",
    "for_each_valid_pfn_and_page",
    "for_each_vma",
    "for_each_vmap_area",
    "in_direct_map",
    "memory_block_size_bytes",
    "mm_cmdline",
    "mm_environ",
    "page_flags",
    "page_index",
    "page_size",
    "page_to_pfn",
    "page_to_phys",
    "page_to_virt",
    "pfn_to_page",
    "pfn_to_virt",
    "phys_to_page",
    "phys_to_virt",
    "task_rss",
    "task_vsize",
    "totalram_pages",
    "virt_to_page",
    "virt_to_pfn",
    "virt_to_phys",
    "vm_commit_limit",
    "vm_memory_committed",
    "vma_find",
    "vma_name",
    "vmalloc_to_page",
    "vmalloc_to_pfn",
    # Generated by scripts/generate_page_flag_getters.py.
    "PageActive",
    "PageChecked",
    "PageDirty",
    "PageDoubleMap",
    "PageError",
    "PageForeign",
    "PageHWPoison",
    "PageHasHWPoisoned",
    "PageIdle",
    "PageIsolated",
    "PageLRU",
    "PageLocked",
    "PageMappedToDisk",
    "PageMlocked",
    "PageOwnerPriv1",
    "PagePinned",
    "PagePrivate",
    "PagePrivate2",
    "PageReadahead",
    "PageReclaim",
    "PageReferenced",
    "PageReported",
    "PageReserved",
    "PageSavePinned",
    "PageSkipKASanPoison",
    "PageSlobFree",
    "PageSwapBacked",
    "PageUncached",
    "PageUnevictable",
    "PageUptodate",
    "PageVmemmapSelfHosted",
    "PageWaiters",
    "PageWorkingset",
    "PageWriteback",
    "PageXenRemapped",
    "PageYoung",
)


def page_flags(page: Object) -> Object:
    """
    Return a page's flags.

    This handles kernel versions before and after the page flags were moved to
    `memdesc_flags_t`.

    :param page: ``struct page *``
    :return: ``unsigned long``
    """
    # Since Linux kernel commit 53fbef56e07d ("mm: introduce memdesc_flags_t")
    # (in v6.18), struct page->flags is no longer an unsigned long, but a
    # memdesc_flags_t containing the actual flags in '.f'.
    try:
        return page.flags.f
    except AttributeError:
        return page.flags


def page_index(page: Object) -> Object:
    """
    Return a page's offset (in pages) within its mapping.

    :param page: ``struct page *``
    :return: ``pgoff_t``
    """
    # The member was renamed in acc53a0b4c15 ("mm: rename page->index to
    # page->__folio_index") (in v6.16).
    try:
        return page.__folio_index
    except AttributeError:
        return page.index


def PageActive(page: Object) -> bool:
    """
    Return whether the ``PG_active`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_active"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageChecked(page: Object) -> bool:
    """
    Return whether the ``PG_checked`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_checked"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageDirty(page: Object) -> bool:
    """
    Return whether the ``PG_dirty`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_dirty"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageDoubleMap(page: Object) -> bool:
    """
    Return whether the ``PG_double_map`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_double_map"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageError(page: Object) -> bool:
    """
    Return whether the ``PG_error`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_error"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageForeign(page: Object) -> bool:
    """
    Return whether the ``PG_foreign`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_foreign"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageHWPoison(page: Object) -> bool:
    """
    Return whether the ``PG_hwpoison`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_hwpoison"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageHasHWPoisoned(page: Object) -> bool:
    """
    Return whether the ``PG_has_hwpoisoned`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_has_hwpoisoned"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageIdle(page: Object) -> bool:
    """
    Return whether the ``PG_idle`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_idle"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageIsolated(page: Object) -> bool:
    """
    Return whether the ``PG_isolated`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_isolated"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageLRU(page: Object) -> bool:
    """
    Return whether the ``PG_lru`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_lru"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageLocked(page: Object) -> bool:
    """
    Return whether the ``PG_locked`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_locked"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageMappedToDisk(page: Object) -> bool:
    """
    Return whether the ``PG_mappedtodisk`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_mappedtodisk"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageMlocked(page: Object) -> bool:
    """
    Return whether the ``PG_mlocked`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_mlocked"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageOwnerPriv1(page: Object) -> bool:
    """
    Return whether the ``PG_owner_priv_1`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_owner_priv_1"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PagePinned(page: Object) -> bool:
    """
    Return whether the ``PG_pinned`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_pinned"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PagePrivate(page: Object) -> bool:
    """
    Return whether the ``PG_private`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_private"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PagePrivate2(page: Object) -> bool:
    """
    Return whether the ``PG_private_2`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_private_2"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageReadahead(page: Object) -> bool:
    """
    Return whether the ``PG_readahead`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_readahead"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageReclaim(page: Object) -> bool:
    """
    Return whether the ``PG_reclaim`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_reclaim"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageReferenced(page: Object) -> bool:
    """
    Return whether the ``PG_referenced`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_referenced"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageReported(page: Object) -> bool:
    """
    Return whether the ``PG_reported`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_reported"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageReserved(page: Object) -> bool:
    """
    Return whether the ``PG_reserved`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_reserved"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageSavePinned(page: Object) -> bool:
    """
    Return whether the ``PG_savepinned`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_savepinned"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageSkipKASanPoison(page: Object) -> bool:
    """
    Return whether the ``PG_skip_kasan_poison`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_skip_kasan_poison"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageSlobFree(page: Object) -> bool:
    """
    Return whether the ``PG_slob_free`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_slob_free"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageSwapBacked(page: Object) -> bool:
    """
    Return whether the ``PG_swapbacked`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_swapbacked"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageUncached(page: Object) -> bool:
    """
    Return whether the ``PG_uncached`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_uncached"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageUnevictable(page: Object) -> bool:
    """
    Return whether the ``PG_unevictable`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_unevictable"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageUptodate(page: Object) -> bool:
    """
    Return whether the ``PG_uptodate`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_uptodate"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageVmemmapSelfHosted(page: Object) -> bool:
    """
    Return whether the ``PG_vmemmap_self_hosted`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_vmemmap_self_hosted"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageWaiters(page: Object) -> bool:
    """
    Return whether the ``PG_waiters`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_waiters"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageWorkingset(page: Object) -> bool:
    """
    Return whether the ``PG_workingset`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_workingset"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageWriteback(page: Object) -> bool:
    """
    Return whether the ``PG_writeback`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_writeback"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageXenRemapped(page: Object) -> bool:
    """
    Return whether the ``PG_xen_remapped`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_xen_remapped"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


def PageYoung(page: Object) -> bool:
    """
    Return whether the ``PG_young`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_young"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))


# End generated by scripts/generate_page_flag_getters.py.


def _get_PageSlab_impl(prog: Program) -> Callable[[Object], bool]:
    # Since Linux kernel commit 46df8e73a4a3 ("mm: free up PG_slab") (in
    # v6.10), slab pages are identified by a page type, which is indicated by a
    # mapcount value matching a value in VMCOREINFO. Before that, they are
    # indicated by a page flag.
    try:
        return prog.cache["PageSlab"]
    except KeyError:
        pass
    vmcoreinfo = prog["VMCOREINFO"].string_()
    match = re.search(
        rb"^NUMBER\(PAGE_SLAB_MAPCOUNT_VALUE\)=(-?[0-9]+)$", vmcoreinfo, flags=re.M
    )
    if match:
        PAGE_SLAB_MAPCOUNT_VALUE = int(match.group(1))

        def PageSlab(page: Object) -> bool:
            return page._mapcount.counter.value_() == PAGE_SLAB_MAPCOUNT_VALUE

    else:
        mask = 1 << prog["PG_slab"]

        def PageSlab(page: Object) -> bool:
            return bool(page_flags(page) & mask)

    prog.cache["PageSlab"] = PageSlab
    return PageSlab


def PageSlab(page: Object) -> bool:
    """
    Return whether a page belongs to the slab allocator.

    :param page: ``struct page *``
    """
    return _get_PageSlab_impl(page.prog_)(page)


# Handle the rename in Linux kernel commit d50569612c29 ("mm: rename the
# 'compound_head' field in the 'struct page' to 'compound_info'") (in v7.1).
def _compound_info_member(prog: Program) -> Optional[str]:
    try:
        return prog.cache["compound_info_member"]
    except KeyError:
        page_type = prog.type("struct page")
        if page_type.has_member("compound_info"):
            member = "compound_info"
        elif page_type.has_member("compound_head"):
            member = "compound_head"
        else:
            member = None
        prog.cache["compound_info_member"] = member
        return member


def PageCompound(page: Object) -> bool:
    """
    Return whether a page is part of a `compound page
    <https://lwn.net/Articles/619514/>`_.

    :param page: ``struct page *``
    """
    prog = page.prog_
    page = page.read_()
    # Since Linux kernel commit 1d798ca3f164 ("mm: make compound_head()
    # robust") (in v4.4), PG_head is always defined, and a tail page has the
    # least significant bit of compound_info/compound_head set. Before that,
    # there is no compound_info/compound_head (and no fake head pages).
    # Instead, if CONFIG_PAGEFLAGS_EXTENDED=y, then PG_head and PG_tail are
    # defined. Otherwise, there is only PG_compound, and PG_reclaim is set for
    # tail pages and clear for head pages.
    try:
        PG_head = prog["PG_head"]
    except KeyError:
        return bool(page_flags(page) & (1 << prog["PG_compound"]))
    else:
        flags = page_flags(page)
        if flags & (1 << PG_head):
            return True
        member = _compound_info_member(prog)
        if member is None:
            return bool(flags & (1 << prog["PG_tail"]))
        else:
            return bool(page.member_(member).value_() & 1)


# Since Linux kernel commit 8c846c879e22 ("mm: rework compound_head() for
# power-of-2 sizeof(struct page)") (in v7.1), compound_info may encode a mask
# instead of a direct pointer when CONFIG_HUGETLB_PAGE_OPTIMIZE_VMEMMAP is
# enabled and sizeof(struct page) is a power of 2. See compound_info_has_mask()
# in the kernel.
def _compound_info_has_mask(prog: Program) -> bool:
    try:
        return prog.cache["compound_info_has_mask"]
    except KeyError:
        page_size = sizeof(prog.type("struct page"))
        has_mask = ((page_size & (page_size - 1)) == 0) and prog.type(
            "struct zone"
        ).has_member("vmemmap_tails")
        prog.cache["compound_info_has_mask"] = has_mask
        return has_mask


# Before Linux kernel commit 8c846c879e22 ("mm: rework compound_head() for
# power-of-2 sizeof(struct page)") (in v7.1), HugeTLB Vmemmap Optimization
# (HVO) creates "fake" head pages that are actually tail pages. See Linux
# kernel commit e7d324850bfc ("mm: hugetlb: free the 2nd vmemmap page
# associated with each HugeTLB page") (in v5.18) and
# https://www.kernel.org/doc/html/latest/mm/vmemmap_dedup.html.
def _page_is_fake_head(page: Object) -> bool:
    head = page[1].compound_head.value_()
    return bool(head & 1) and (head - 1) != page.value_()


def PageHead(page: Object) -> bool:
    """
    Return whether a page is a head page in a `compound page`_.

    :param page: ``struct page *``
    """
    prog = page.prog_
    page = page.read_()
    # See PageCompound() re: Linux kernel commit 1d798ca3f164 ("mm: make
    # compound_head() robust") (in v4.4).
    try:
        PG_head = prog["PG_head"]
    except KeyError:
        PG_head_mask = 1 << prog["PG_compound"]
        PG_head_tail_mask = PG_head_mask | (1 << prog["PG_reclaim"])
        return (page_flags(page) & PG_head_tail_mask) == PG_head_mask
    else:
        return bool(page_flags(page) & (1 << PG_head)) and not (
            _compound_info_member(prog) == "compound_head" and _page_is_fake_head(page)
        )


def PageTail(page: Object) -> bool:
    """
    Return whether a page is a tail page in a `compound page`_.

    :param page: ``struct page *``
    """
    prog = page.prog_
    page = page.read_()
    # See PageCompound() re: Linux kernel commit 1d798ca3f164 ("mm: make
    # compound_head() robust") (in v4.4).
    member = _compound_info_member(prog)
    if member is None:
        try:
            PG_tail = prog["PG_tail"]
        except KeyError:
            PG_head_tail_mask = (1 << prog["PG_compound"]) | (1 << prog["PG_reclaim"])
            return (page_flags(page) & PG_head_tail_mask) == PG_head_tail_mask
        else:
            return bool(page_flags(page) & (1 << PG_tail))
    if page.member_(member).value_() & 1:
        return True
    return (
        member == "compound_head"
        and bool(page_flags(page) & (1 << prog["PG_head"]))
        and _page_is_fake_head(page)
    )


def compound_head(page: Object) -> Object:
    """
    Get the head page associated with a page.

    If *page* is a tail page, this returns the head page of the `compound
    page`_ it belongs to. Otherwise, it returns *page*.

    :param page: ``struct page *``
    :return: ``struct page *``
    """
    prog = page.prog_
    page = page.read_()
    member = _compound_info_member(prog)
    if member is None:
        # Before Linux kernel commit 1d798ca3f164 ("mm: make compound_head()
        # robust") (in v4.4), the head page is in page->first_page, and there
        # are no fake head pages.
        return page.first_page.read_() if PageTail(page) else page
    info = page.member_(member).read_()
    if info & 1:
        if member == "compound_info" and _compound_info_has_mask(prog):
            return Object(prog, page.type_, page.value_() & info.value_())
        else:
            return cast(page.type_, info - 1)
    # Handle fake head pages (see _page_is_fake_head()).
    if member == "compound_head" and (page_flags(page) & (1 << prog["PG_head"])):
        head = page[1].compound_head.read_()
        if head & 1:
            return cast(page.type_, head - 1)
    return page


def compound_order(page: Object) -> Object:
    """
    Return the allocation order of a potentially `compound page`_.

    :param page: ``struct page *``
    :return: ``unsigned int``
    """
    prog = page.prog_

    if not PageHead(page):
        return Object(prog, "unsigned int", 0)

    # Since Linux kernel commit ebc1baf5c9b4 ("mm: free up a word in the first
    # tail page") (in v6.6), the compound order is in the low byte of struct
    # folio::_flags_1 (from_folio = 2). Between that and Linux kernel commit
    # Linux kernel commit 379708ffde1b ("mm: add the first tail page to struct
    # folio") (in v6.1), the compound order is in struct folio::_folio_order
    # (from_folio = 1). Before Linux kernel commit 1c5509be58f6 ("mm: remove
    # 'First tail page' members from struct page") (in v6.3), the compound
    # order is in struct page::compound_order of the first tail page
    # (from_folio = 0).
    try:
        from_folio = prog.cache["compound_order_from_folio"]
    except KeyError:
        from_folio = 0
        try:
            struct_folio = prog.type("struct folio")
        except LookupError:
            pass
        else:
            if struct_folio.has_member("_folio_order"):
                from_folio = 1
            elif struct_folio.has_member("_flags_1"):
                from_folio = 2
        prog.cache["compound_order_from_folio"] = from_folio
    if from_folio == 2:
        return cast("unsigned int", cast("struct folio *", page)._flags_1 & 0xFF)
    elif from_folio == 1:
        return cast("unsigned int", cast("struct folio *", page)._folio_order)
    else:
        return cast("unsigned int", page[1].compound_order)


def compound_nr(page: Object) -> Object:
    """
    Return the number of pages in a potentially `compound page`_.

    :param page: ``struct page *``
    :return: ``unsigned long``
    """
    return Object(page.prog_, "unsigned long", 1) << compound_order(page)


def page_size(page: Object) -> Object:
    """
    Return the number of bytes in a potentially `compound page`_.

    :param page: ``struct page *``
    :return: ``unsigned long``
    """
    return page.prog_["PAGE_SIZE"] << compound_order(page)


def decode_page_flags(page: Object) -> str:
    """
    Get a human-readable representation of the flags set on a page.

    >>> decode_page_flags(page)
    'PG_uptodate|PG_dirty|PG_lru|PG_reclaim|PG_swapbacked|PG_readahead|PG_savepinned|PG_isolated|PG_reported'

    :param page: ``struct page *``
    """
    return decode_page_flags_value(page_flags(page))


@takes_program_or_default
def decode_page_flags_value(prog: Program, flags: IntegerLike) -> str:
    """
    Get a human-readable representation of the flags value from a page.

    >>> flags = page_flags(page).read_()
    >>> hex(flags)
    0xfffffd0004028
    >>> decode_page_flags_value(flags)
    'PG_uptodate|PG_lru|PG_private|PG_reported'

    See also the :func:`decode_page_flags()` shortcut, which takes a
    ``struct page *`` instead.

    :param flags: ``unsigned long``
    """
    NR_PAGEFLAGS = prog["__NR_PAGEFLAGS"]
    PAGEFLAGS_MASK = (1 << NR_PAGEFLAGS.value_()) - 1
    return decode_enum_type_flags(
        operator.index(flags) & PAGEFLAGS_MASK, NR_PAGEFLAGS.type_
    )


# Get the struct page * for PFN 0.
def _page0(prog: Program) -> Object:
    try:
        return prog.cache["page0"]
    except KeyError:
        pass
    try:
        # With CONFIG_SPARSEMEM_VMEMMAP=y, page 0 is vmemmap.
        page0 = prog["vmemmap"]
    except KeyError:
        page0 = prog["mem_map"] - prog["ARCH_PFN_OFFSET"]
    # The struct page array is not contiguous for CONFIG_SPARSEMEM=y with
    # CONFIG_SPARSEMEM_VMEMMAP=n or CONFIG_DISCONTIGMEM=y, so those are not
    # supported yet.
    prog.cache["page0"] = page0
    return page0


@takes_program_or_default
def for_each_page(prog: Program) -> Iterator[Object]:
    """
    Iterate over every valid ``struct page *``.

    :return: Iterator of ``struct page *`` objects.
    """
    for start_pfn, end_pfn, mem_map in for_each_valid_page_range(prog):
        for pfn in range(start_pfn, end_pfn):
            yield mem_map + pfn


@takes_program_or_default
def for_each_valid_pfn_and_page(prog: Program) -> Iterator[Tuple[int, Object]]:
    """
    Iterate over every valid page frame number (PFN) and ``struct page *``.

    :return: Iterator of (``pfn``, ``struct page *``) tuples.
    """
    for start_pfn, end_pfn, mem_map in for_each_valid_page_range(prog):
        for pfn in range(start_pfn, end_pfn):
            yield pfn, mem_map + pfn


def _for_each_valid_page_range_flatmem(
    prog: Program,
) -> Iterator[Tuple[int, int, Object]]:
    mem_map = _page0(prog)

    if (
        prog.platform.arch  # type: ignore[union-attr]  # platform can't be None
        == Architecture.ARM
    ):
        # Since Linux kernel commit a4d5613c4dc6 ("arm: extend pfn_valid to
        # take into account freed memory map alignment") (in v5.14), Arm's
        # pfn_valid() checks that the PFN lies within a pageblock that
        # intersects a present memory chunk. However, pageblock_nr_pages is a
        # macro, and it's not easy to get its value. So, we get as close as we
        # can and ignore the extra pages granted by the pageblock alignment
        # (which is also what the kernel did before Linux kernel commit
        # 09414d00a137 ("ARM: only consider memblocks with NOMAP cleared for
        # linear mapping") (in v4.5)).
        page_shift = prog["PAGE_SHIFT"].value_()
        memory = prog["memblock"].memory
        prev_start = prev_end = None
        for region in memory.regions[: memory.cnt]:
            start = region.base.value_()
            end = (start + region.size.value_()) >> page_shift
            start >>= page_shift
            if start == prev_end:
                prev_end = end  # Merge adjacent regions.
            else:
                if prev_start is not None:
                    yield prev_start, prev_end, mem_map
                prev_start = start
                prev_end = end
        if prev_start is not None:
            yield prev_start, prev_end, mem_map  # type: ignore  # prev_end can't be None
        return

    # Generic FLATMEM validity.
    start_pfn = prog["ARCH_PFN_OFFSET"].value_()
    yield start_pfn, start_pfn + prog["max_mapnr"].value_(), mem_map


@takes_program_or_default
def for_each_valid_page_range(prog: Program) -> Iterator[Tuple[int, int, Object]]:
    """
    Iterate over every contiguous range of valid page frame numbers and
    ``struct page``\\ s.

    >>> for start_pfn, end_pfn, mem_map in for_each_valid_page():
    ...     pages = mem_map[start_pfn:end_pfn]

    :return: Iterator of (``start_pfn``, ``end_pfn``, ``mem_map``) tuples.
        ``start_pfn`` is the minimum page frame number (PFN) in the range
        (inclusive). ``end_pfn`` is the maximum PFN in the range (exclusive).
        ``mem_map`` is a ``struct page *`` object such that ``mem_map[pfn]`` is
        the ``struct page`` for the given PFN.
    """
    try:
        mem_section = prog["mem_section"]
    except ObjectNotFoundError:
        yield from _for_each_valid_page_range_flatmem(prog)
        return

    # To support SPARSEMEM without SPARSEMEM_VMEMMAP, we will need to check
    # whether each section's mem_map is contiguous.
    mem_map = prog["vmemmap"].read_()

    PAGE_SHIFT = prog["PAGE_SHIFT"].value_()
    SECTIONS_PER_ROOT = prog["SECTIONS_PER_ROOT"].value_()
    SECTION_SIZE_BITS = prog["SECTION_SIZE_BITS"].value_()
    PAGES_PER_SECTION = 1 << (SECTION_SIZE_BITS - PAGE_SHIFT)
    SUBSECTION_SHIFT = 21
    SUBSECTIONS_PER_SECTION = 1 << (SECTION_SIZE_BITS - SUBSECTION_SHIFT)
    PAGES_PER_SUBSECTION = 1 << (SUBSECTION_SHIFT - PAGE_SHIFT)
    flags = _section_flags(prog)
    SECTION_HAS_MEM_MAP = flags["SECTION_HAS_MEM_MAP"]
    SECTION_IS_EARLY = flags["SECTION_IS_EARLY"]

    highest_present_section_nr = _highest_present_section_nr(prog)
    nr_roots = highest_present_section_nr // SECTIONS_PER_ROOT + 1

    unaliased_type = mem_section.type_.unaliased()
    if unaliased_type.kind == TypeKind.POINTER:
        mem_section = mem_section.read_()
        if not mem_section:
            return

    root_kind = unaliased_type.type.unaliased_kind()

    pfn = 0
    start_pfn = None
    for root_nr, root in enumerate(mem_section[:nr_roots]):
        if root_kind == TypeKind.POINTER:
            root = root.read_()
            if not root:
                if start_pfn is not None:
                    yield start_pfn, pfn, mem_map
                    start_pfn = None
                pfn += SECTIONS_PER_ROOT * PAGES_PER_SECTION
                continue

        if root_nr == nr_roots - 1:
            nr_sections = highest_present_section_nr % SECTIONS_PER_ROOT + 1
        else:
            nr_sections = SECTIONS_PER_ROOT
        for section in root[:nr_sections]:
            mem_map_value = section.section_mem_map.value_()
            # Open-coded valid_section() and early_section() to avoid some
            # overhead.
            if mem_map_value & SECTION_HAS_MEM_MAP:
                # struct mem_section::usage only exists since Linux kernel
                # commit f1eca35a0dc7 ("mm/sparsemem: introduce struct
                # mem_section_usage") (in v5.3). Additionally, struct
                # mem_section_usage::subsection_map only exists for
                # CONFIG_SPARSEMEM_VMEMMAP. Without both, as well as for early
                # sections, validity has section granularity.
                subsection_map = None
                if not (mem_map_value & SECTION_IS_EARLY):
                    try:
                        subsection_map = section.usage.subsection_map
                    except AttributeError:
                        pass

                if subsection_map is None:
                    if start_pfn is None:
                        start_pfn = pfn
                else:
                    end_bit = None if start_pfn is None else 0
                    for bit in for_each_set_bit(
                        subsection_map, SUBSECTIONS_PER_SECTION
                    ):
                        if bit != end_bit:
                            if start_pfn is not None:
                                yield start_pfn, pfn + end_bit * PAGES_PER_SUBSECTION, mem_map
                            start_pfn = pfn + bit * PAGES_PER_SUBSECTION
                        end_bit = bit + 1
                    if end_bit != SUBSECTIONS_PER_SECTION and start_pfn is not None:
                        yield start_pfn, pfn + end_bit * PAGES_PER_SUBSECTION, mem_map
                        start_pfn = None
            elif start_pfn is not None:
                yield start_pfn, pfn, mem_map
                start_pfn = None
            pfn += PAGES_PER_SECTION

    if start_pfn is not None:
        yield start_pfn, pfn, mem_map


@takes_program_or_default
def PFN_PHYS(prog: Program, pfn: IntegerLike) -> Object:
    """
    Get the physical address of a page frame number (PFN).

    :param pfn: ``unsigned long``
    :return: ``phys_addr_t``
    """
    return Object(prog, "phys_addr_t", pfn) << prog["PAGE_SHIFT"]


@takes_program_or_default
def PHYS_PFN(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the page frame number (PFN) of a physical address.

    :param addr: ``phys_addr_t``
    :return: ``unsigned long``
    """
    return Object(prog, "unsigned long", addr) >> prog["PAGE_SHIFT"]


def page_to_pfn(page: Object) -> Object:
    """
    Get the page frame number (PFN) of a page.

    :param page: ``struct page *``
    :return: ``unsigned long``
    """
    return cast("unsigned long", page - _page0(page.prog_))


def page_to_phys(page: Object) -> Object:
    """
    Get the physical address of a page.

    :param page: ``struct page *``
    :return: ``phys_addr_t``
    """
    return PFN_PHYS(page_to_pfn(page))


def page_to_virt(page: Object) -> Object:
    """
    Get the directly mapped virtual address of a page.

    :param page: ``struct page *``
    :return: ``void *``
    """
    return pfn_to_virt(page_to_pfn(page))


@takes_program_or_default
def pfn_to_page(prog: Program, pfn: IntegerLike) -> Object:
    """
    Get the page with a page frame number (PFN).

    :param pfn: ``unsigned long``
    :return: ``struct page *``
    """
    return _page0(prog) + pfn


@takes_program_or_default
def pfn_to_virt(prog: Program, pfn: IntegerLike) -> Object:
    """
    Get the directly mapped virtual address of a page frame number (PFN).

    :param pfn: ``unsigned long``
    :return: ``void *``
    """
    return phys_to_virt(PFN_PHYS(prog, pfn))


@takes_program_or_default
def phys_to_page(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the page containing a physical address.

    :param addr: ``phys_addr_t``
    :return: ``struct page *``
    """
    return pfn_to_page(PHYS_PFN(prog, addr))


@takes_program_or_default
def phys_to_virt(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the directly mapped virtual address of a physical address.

    :param addr: ``phys_addr_t``
    :return: ``void *``
    """
    return Object(
        prog, "void *", operator.index(addr) + _linux_helper_direct_mapping_offset(prog)
    )


@takes_program_or_default
def virt_to_page(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the page containing a directly mapped virtual address.

    .. _mm-helpers-direct-map:

    .. note::

        This only works for virtual addresses from the "direct map". This
        includes address from:

        * kmalloc
        * Slab allocator
        * Page allocator

        But not:

        * vmalloc
        * vmap
        * ioremap
        * Symbols (function pointers, global variables)

        For vmalloc or vmap addresses, use :func:`vmalloc_to_page(addr)
        <vmalloc_to_page>`. For arbitrary kernel addresses, use
        :func:`follow_page(prog["init_mm"].address_of_(), addr) <follow_page>`.

    :param addr: ``void *``
    :return: ``struct page *``
    """
    return pfn_to_page(virt_to_pfn(prog, addr))


@takes_program_or_default
def virt_to_pfn(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the page frame number (PFN) of a directly mapped virtual address.

    .. note::

        This only works for virtual addresses from the :ref:`"direct map"
        <mm-helpers-direct-map>`. For vmalloc or vmap addresses, use
        :func:`vmalloc_to_pfn(addr) <vmalloc_to_pfn>`. For arbitrary kernel
        addresses, use :func:`follow_pfn(prog["init_mm"].address_of_(), addr)
        <follow_pfn>`.

    :param addr: ``void *``
    :return: ``unsigned long``
    """
    return PHYS_PFN(virt_to_phys(prog, addr))


@takes_program_or_default
def virt_to_phys(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the physical address of a directly mapped virtual address.

    .. note::

        This only works for virtual addresses from the :ref:`"direct map"
        <mm-helpers-direct-map>`. For arbitrary kernel addresses, use
        :func:`follow_phys(prog["init_mm"].address_of_(), addr) <follow_phys>`.

    :param addr: ``void *``
    :return: ``phys_addr_t``
    """
    return Object(
        prog,
        "unsigned long",
        operator.index(addr) - _linux_helper_direct_mapping_offset(prog),
    )


def follow_page(mm: Object, addr: IntegerLike) -> Object:
    """
    Get the page that a virtual address maps to in a virtual address space.

    >>> task = find_task(113)
    >>> follow_page(task.mm, 0x7fffbbb6d4d0)
    *(struct page *)0xffffbe4bc0337b80 = {
        ...
    }

    :param mm: ``struct mm_struct *``
    :param addr: ``void *``
    :return: ``struct page *``
    :raises FaultError: if the virtual address
        :ref:`cannot be translated <virtual address translation failures>`
    :raises NotImplementedError: if virtual address translation is :ref:`not
        supported <architecture support matrix>` for this architecture yet
    """
    return phys_to_page(follow_phys(mm, addr))


def follow_pfn(mm: Object, addr: IntegerLike) -> Object:
    """
    Get the page frame number (PFN) that a virtual address maps to in a virtual
    address space.

    >>> task = find_task(113)
    >>> follow_pfn(task.mm, 0x7fffbbb6d4d0)
    (unsigned long)52718

    :param mm: ``struct mm_struct *``
    :param addr: ``void *``
    :return: ``unsigned long``
    :raises FaultError: if the virtual address
        :ref:`cannot be translated <virtual address translation failures>`
    :raises NotImplementedError: if virtual address translation is :ref:`not
        supported <architecture support matrix>` for this architecture yet
    """
    return PHYS_PFN(follow_phys(mm, addr))


def follow_phys(mm: Object, addr: IntegerLike) -> Object:
    """
    Get the physical address that a virtual address maps to in a virtual
    address space.

    >>> task = find_task(113)
    >>> follow_phys(task.mm, 0x7fffbbb6d4d0)
    (phys_addr_t)215934160

    :param mm: ``struct mm_struct *``
    :param addr: ``void *``
    :return: ``phys_addr_t``
    :raises FaultError: if the virtual address
        :ref:`cannot be translated <virtual address translation failures>`
    :raises NotImplementedError: if virtual address translation is :ref:`not
        supported <architecture support matrix>` for this architecture yet
    """
    prog = mm.prog_
    return Object(prog, "phys_addr_t", _linux_helper_follow_phys(prog, mm.pgd, addr))


@takes_program_or_default
def vmalloc_to_page(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the page containing a vmalloc or vmap address.

    >>> task = find_task(113)
    >>> vmalloc_to_page(task.stack)
    *(struct page *)0xffffbe4bc00a2200 = {
        ...
    }

    :param addr: ``void *``
    :return: ``struct page *``
    """
    return follow_page(prog["init_mm"].address_of_(), addr)


@takes_program_or_default
def vmalloc_to_pfn(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the page frame number (PFN) containing a vmalloc or vmap address.

    >>> task = find_task(113)
    >>> vmalloc_to_pfn(task.stack)
    (unsigned long)10376

    :param addr: ``void *``
    :return: ``unsigned long``
    """
    return page_to_pfn(vmalloc_to_page(prog, addr))


def _vmap_area_rb_cmp(addr: int, va: Object) -> int:
    if addr < va.va_start.value_():
        return -1
    elif addr >= va.va_end.value_():
        return 1
    else:
        return 0


# On !SMP or 32-bit kernels, vmap_nodes, nr_vmap_nodes, and vmap_zone_size can
# only by &single, 1, and 1, respectively. In that case, GCC as of version 12.2
# sometimes doesn't generate location descriptions for those variables.
def _vmap_nodes(prog: Program) -> Object:
    vmap_nodes = prog["vmap_nodes"]
    try:
        return vmap_nodes.read_()
    except ObjectAbsentError:
        return prog.variable("single", "mm/vmalloc.c").address_of_()


def _nr_vmap_nodes(prog: Program) -> int:
    try:
        return prog["nr_vmap_nodes"].value_()
    except ObjectAbsentError:
        return 1


def _vmap_zone_size(prog: Program) -> int:
    try:
        return prog["vmap_zone_size"].value_()
    except ObjectAbsentError:
        return 1


@takes_program_or_default
def find_vmap_area(prog: Program, addr: IntegerLike) -> Object:
    """
    Return the ``struct vmap_area *`` containing an address.

    >>> find_vmap_area(0xffffa2b680081000)
    *(struct vmap_area *)0xffffa16541046b40 = {
            ...
    }

    :param addr: Address to look up.
    :return: ``struct vmap_area *`` (``NULL`` if not found)
    """
    addr = operator.index(addr)
    # Since Linux kernel commit d093602919ad ("mm: vmalloc: remove global
    # vmap_area_root rb-tree") (in v6.9), vmap areas are split up in multiple
    # red-black trees in separate "nodes". Before that, they're in a single
    # red-black tree.
    try:
        vmap_nodes = _vmap_nodes(prog)
    except KeyError:
        return rb_find(
            "struct vmap_area",
            prog["vmap_area_root"].address_of_(),
            "rb_node",
            addr,
            _vmap_area_rb_cmp,
        )
    else:
        nr_vmap_nodes = _nr_vmap_nodes(prog)
        i = j = (addr // _vmap_zone_size(prog)) % nr_vmap_nodes
        while True:
            vn = vmap_nodes[i]
            va = rb_find(
                "struct vmap_area",
                vn.busy.root.address_of_(),
                "rb_node",
                addr,
                _vmap_area_rb_cmp,
            )
            if va:
                return va
            # As noted in the kernel implementation, the given address may be
            # in a different node than the start address, so we have to loop.
            i = (i + 1) % nr_vmap_nodes
            if i == j:
                return NULL(prog, "struct vmap_area *")


@takes_program_or_default
def for_each_vmap_area(prog: Program) -> Iterator[Object]:
    """
    Iterate over every ``struct vmap_area *`` on the system.

    >>> for va in for_each_vmap_area():
    ...     caller = ""
    ...     if va.vm:
    ...         try:
    ...             sym = prog.symbol(va.vm.caller)
    ...         except LookupError:
    ...             pass
    ...         else:
    ...             caller = f" {sym.name}"
    ...     print(f"{hex(va.va_start)}-{hex(va.va_end)}{caller}")
    ...
    0xffffa2b680000000-0xffffa2b680005000 irq_init_percpu_irqstack
    0xffffa2b680005000-0xffffa2b680007000 acpi_os_map_iomem
    0xffffa2b68000b000-0xffffa2b68000d000 hpet_enable
    0xffffa2b680080000-0xffffa2b680085000 kernel_clone
    ...

    :return: Iterator of ``struct vmap_area *`` objects.
    """
    # Since Linux kernel commit d093602919ad ("mm: vmalloc: remove global
    # vmap_area_root rb-tree") (in v6.9), vmap areas are split up in multiple
    # lists in separate "nodes". Before that, they're in a single list.
    try:
        vmap_nodes = _vmap_nodes(prog)
    except KeyError:
        yield from list_for_each_entry(
            "struct vmap_area", prog["vmap_area_list"].address_of_(), "list"
        )
    else:
        for i in range(_nr_vmap_nodes(prog)):
            yield from list_for_each_entry(
                "struct vmap_area", vmap_nodes[i].busy.head.address_of_(), "list"
            )


def access_process_vm(task: Object, address: IntegerLike, size: IntegerLike) -> bytes:
    """
    Read memory from a task's virtual address space.

    >>> task = find_task(1490152)
    >>> access_process_vm(task, 0x7f8a62b56da0, 12)
    b'hello, world'

    :param task: ``struct task_struct *``
    :param address: Starting address.
    :param size: Number of bytes to read.
    :raises FaultError: if the virtual address
        :ref:`cannot be translated <virtual address translation failures>`
    :raises NotImplementedError: if virtual address translation is :ref:`not
        supported <architecture support matrix>` for this architecture yet
    """
    return _linux_helper_read_vm(task.prog_, task.mm.pgd, address, size)


def access_remote_vm(mm: Object, address: IntegerLike, size: IntegerLike) -> bytes:
    """
    Read memory from a virtual address space. This is similar to
    :func:`access_process_vm()`, but it takes a ``struct mm_struct *`` instead
    of a ``struct task_struct *``.

    >>> task = find_task(1490152)
    >>> access_remote_vm(task.mm, 0x7f8a62b56da0, 12)
    b'hello, world'

    :param mm: ``struct mm_struct *``
    :param address: Starting address.
    :param size: Number of bytes to read.
    :raises FaultError: if the virtual address
        :ref:`cannot be translated <virtual address translation failures>`
    :raises NotImplementedError: if virtual address translation is :ref:`not
        supported <architecture support matrix>` for this architecture yet
    """
    return _linux_helper_read_vm(mm.prog_, mm.pgd, address, size)


def cmdline(task: Object) -> Optional[List[bytes]]:
    """
    Get the list of command line arguments of a task, or ``None`` for kernel tasks.

    >>> cmdline(find_task(1495216))
    [b'vim', b'drgn/helpers/linux/mm.py']

    .. code-block:: console

        $ tr '\\0' ' ' < /proc/1495216/cmdline
        vim drgn/helpers/linux/mm.py

    :param task: ``struct task_struct *``
    :raises FaultError: if the virtual address containing the command line
        :ref:`cannot be translated <virtual address translation failures>`
    :raises NotImplementedError: if virtual address translation is :ref:`not
        supported <architecture support matrix>` for this architecture yet
    """
    mm = task.mm.read_()
    return mm_cmdline(mm) if mm else None


def mm_cmdline(mm: Object) -> List[bytes]:
    """
    Like :func:`cmdline()`, but takes a (non-``NULL``) ``struct mm_struct *``
    instead of a ``struct task_struct *``.

    :param mm: ``struct mm_struct *``
    """
    mm = mm.read_()
    arg_start = mm.arg_start.value_()
    arg_end = mm.arg_end.value_()
    return access_remote_vm(mm, arg_start, arg_end - arg_start).split(b"\0")[:-1]


def environ(task: Object) -> Optional[List[bytes]]:
    """
    Get the list of environment variables of a task, or ``None`` for kernel tasks.

    >>> environ(find_task(1497797))
    [b'HOME=/root', b'PATH=/usr/local/sbin:/usr/local/bin:/usr/bin', b'LOGNAME=root']

    .. code-block:: console

        $ tr '\\0' '\\n' < /proc/1497797/environ
        HOME=/root
        PATH=/usr/local/sbin:/usr/local/bin:/usr/bin
        LOGNAME=root

    :param task: ``struct task_struct *``
    :raises FaultError: if the virtual address containing the environment
        :ref:`cannot be translated <virtual address translation failures>`
    :raises NotImplementedError: if virtual address translation is :ref:`not
        supported <architecture support matrix>` for this architecture yet
    """
    mm = task.mm.read_()
    return mm_environ(mm) if mm else None


def mm_environ(mm: Object) -> List[bytes]:
    """
    Like :func:`environ()`, but takes a (non-``NULL``) ``struct mm_struct *``
    instead of a ``struct task_struct *``.

    :param mm: ``struct mm_struct *``
    """
    mm = mm.read_()
    env_start = mm.env_start.value_()
    env_end = mm.env_end.value_()
    return access_remote_vm(mm, env_start, env_end - env_start).split(b"\0")[:-1]


def _vma_rb_cmp(addr: int, vma: Object) -> int:
    if addr < vma.vm_start.value_():
        return -1
    elif addr >= vma.vm_end.value_():
        return 1
    else:
        return 0


def vma_find(mm: Object, addr: IntegerLike) -> Object:
    """
    Return the virtual memory area (VMA) containing an address.

    :param mm: ``struct mm_struct *``
    :param addr: Address to look up.
    :return: ``struct vm_area_struct *`` (``NULL`` if not found)
    """
    try:
        # Since Linux kernel commit 524e00b36e8c ("mm: remove rb tree.") (in
        # v6.1), VMAs are stored in a maple tree.
        mt = mm.mm_mt.address_of_()
    except AttributeError:
        # Before that, they are in a red-black tree.
        return rb_find(
            "struct vm_area_struct",
            mm.mm_rb.address_of_(),
            "vm_rb",
            operator.index(addr),
            _vma_rb_cmp,
        )
    else:
        return cast("struct vm_area_struct *", mtree_load(mt, addr))


def vma_name(vma: Object) -> bytes:
    """
    Get the display name or file path for a Virtual Memory Area (VMA).

    This helper returns a human-readable label for a given
    ``struct vm_area_struct *`` object, describing what type of memory region
    it represents. It mimics the VMA naming behavior seen in
    ``/proc/<pid>/maps``.

    :param vma: ``struct vm_area_struct *``
    :return: File path or descriptive name of the VMA region. Possible return
        values include:

        - **File path** (e.g., ``b"/usr/lib/libc.so.6"``):
          For VMAs backed by files (``vma->vm_file`` non-NULL).
        - ``b"[heap]"``:
          Heap segment, between ``mm->start_brk`` and ``mm->brk``.
        - ``b"[stack]"``:
          Stack segment containing ``mm->start_stack``.
        - ``b"[vdso]"``:
          Virtual Dynamic Shared Object area (``mm->context.vdso``).
        - ``b"[vvar]"``, ``b"[vsyscall]"``, etc.:
          Architecture-specific special mappings from ``vm_ops->name``.
        - ``b"[anon:<name>]"``: private anonymous memory with a name set by
          `PR_SET_VMA_ANON_NAME
          <https://man7.org/linux/man-pages/man2/PR_SET_VMA.2const.html>`_.
        - ``b"[anon_shmem:<name>]"``: shared memory with a name set by
          `PR_SET_VMA_ANON_NAME
          <https://man7.org/linux/man-pages/man2/PR_SET_VMA.2const.html>`_.
        - Empty string: anonymous memory.
    """
    prog = vma.prog_
    mm = vma.vm_mm.read_()

    vm_file = vma.vm_file.read_()
    if vm_file:
        # Since Linux kernel commit d09e8ca6cb93 ("mm: anonymous shared memory
        # naming") (in v6.2), vma->anon_name exists and is valid iff
        # CONFIG_ANON_VMA_NAME=y. Between that commit and commit 9a10064f5625
        # ("mm: add a field to store names for private anonymous memory") (in
        # v5.17), it exists in a union with vma->shared, so it is only valid
        # when vma->vm_file is NULL. Before that, it doesn't exist.
        if mm:
            try:
                anon_name = vma.anon_name
            except AttributeError:
                pass
            else:
                if anon_name.address_ != vma.shared.address_:
                    anon_name = anon_name.read_()
                    if anon_name:
                        return b"[anon_shmem:" + anon_name.name.string_() + b"]"

        return d_path(vm_file.f_path)

    vm_ops = vma.vm_ops.read_()
    if vm_ops:
        vm_ops_name = vm_ops.name.read_()
        if vm_ops_name == prog["special_mapping_name"]:
            return cast(
                "struct vm_special_mapping *", vma.vm_private_data
            ).name.string_()

    if not mm:
        return b"[vdso]"

    start = vma.vm_start.value_()

    # Before Linux kernel commits c1bab64360e6 ("powerpc/vdso: Move to
    # _install_special_mapping() and remove arch_vma_name()") and c1bab64360e6
    # ("powerpc/vdso: Move to _install_special_mapping() and remove
    # arch_vma_name()") (in v5.11), PowerPC needs a special case for the vdso.
    if prog.platform.arch == Architecture.PPC64:  # type: ignore[union-attr]  # platform can't be None.
        try:
            vdso_base = mm.context.vdso_base
        except AttributeError:
            pass
        else:
            if start == vdso_base.value_():
                return b"[vdso]"

    end = vma.vm_end.value_()

    if (
        mm.start_brk
        and mm.brk
        and start <= mm.brk.value_()
        and end >= mm.start_brk.value_()
    ):
        return b"[heap]"

    if mm.start_stack and start <= mm.start_stack.value_() <= end:
        return b"[stack]"

    # See above regarding the existence of vma->anon_name.
    try:
        anon_name = vma.anon_name.read_()
    except AttributeError:
        pass
    else:
        if anon_name:
            return b"[anon:" + anon_name.name.string_() + b"]"

    return b""


def for_each_vma(mm: Object) -> Iterator[Object]:
    """
    Iterate over every virtual memory area (VMA) in a virtual address space.

    >>> for vma in for_each_vma(task.mm):
    ...     print(vma)
    ...
    *(struct vm_area_struct *)0xffff97ad82bfc930 = {
        ...
    }
    *(struct vm_area_struct *)0xffff97ad82bfc0a8 = {
        ...
    }
    ...

    :param mm: ``struct mm_struct *``
    :return: Iterator of ``struct vm_area_struct *`` objects.
    """
    try:
        # Since Linux kernel commit 763ecb035029 ("mm: remove the vma linked
        # list") (in v6.1), VMAs are stored in a maple tree.
        mt = mm.mm_mt.address_of_()
    except AttributeError:
        # Before that, they are in a linked list.
        vma = mm.mmap
        while vma:
            yield vma
            vma = vma.vm_next
    else:
        type = mm.prog_.type("struct vm_area_struct *")
        for _, _, entry in mt_for_each(mt):
            yield cast(type, entry)


@takes_program_or_default
def totalram_pages(prog: Program) -> int:
    """Return the total number of RAM pages."""
    try:
        # The variable is present since Linux kernel commit ca79b0c211af63fa32
        # ("mm: convert totalram_pages and totalhigh_pages variables
        # to atomic") (in v5.0).
        return prog["_totalram_pages"].counter.value_()
    except KeyError:
        return prog["totalram_pages"].value_()


@takes_program_or_default
def vm_commit_limit(prog: Program) -> int:
    """Get the limit on committed virtual address space in pages."""
    sysctl_overcommit_kbytes = prog["sysctl_overcommit_kbytes"].value_()
    if sysctl_overcommit_kbytes:
        allowed = sysctl_overcommit_kbytes >> (prog["PAGE_SHIFT"].value_() - 10)
    else:
        # Avoid circular import.
        from drgn.helpers.linux.hugetlb import hugetlb_total_pages

        allowed = (
            (totalram_pages(prog) - hugetlb_total_pages(prog))
            * prog["sysctl_overcommit_ratio"].value_()
            // 100
        )
    return allowed + prog["total_swap_pages"].value_()


@takes_program_or_default
def vm_memory_committed(prog: Program) -> int:
    """Get the number of pages of committed virtual address space."""
    return percpu_counter_sum_positive(prog["vm_committed_as"].address_of_())


class PageUsage(NamedTuple):
    """Memory usage statistic in page units."""

    pages: int
    """Total number of pages."""

    free_pages: int
    """Number of free pages."""

    @property
    def used_pages(self) -> int:
        """Number of used pages."""
        return self.pages - self.free_pages


@takes_program_or_default
def in_direct_map(prog: Program, addr: IntegerLike) -> bool:
    """
    Return whether an address is within the kernel's direct memory mapping.

    :param addr: address to check
    """
    addr = operator.index(addr)
    start_addr = pfn_to_virt(prog["min_low_pfn"]).value_()
    end_addr = (pfn_to_virt(prog["max_low_pfn"]) + prog["PAGE_SIZE"]).value_()
    return start_addr <= addr < end_addr


class TaskRss(NamedTuple):
    """Task's resident set size returned by :func:`task_rss()`."""

    file: int
    """Number of resident file pages."""
    anon: int
    """Number of resident anonymous pages."""
    shmem: int
    """Number of resident shared memory pages."""
    swap: int
    """Number of swapped-out anonymous private pages."""

    @property
    def total(self) -> int:
        """
        Total number of resident pages (:attr:`file` + :attr:`anon` +
        :attr:`shmem`).
        """
        return self.file + self.anon + self.shmem


@takes_program_or_default
def task_rss(prog: Program, task: Object) -> TaskRss:
    """
    Return a task's resident set size (RSS) in pages.

    The task's RSS is the number of pages which are currently resident in
    memory. The RSS values can be broken down into anonymous pages (not bound
    to any file), file pages (those associated with memory mapped files), and
    shared memory pages (those which aren't associated with on-disk files, but
    belonging to shared memory mappings). This function returns a named tuple
    containing each category, but the common behavior is to use the
    :attr:`~TaskRss.total` value which sums them up.

    :param task: ``struct task_struct *``
    """

    mm = task.mm.read_()

    # Kthreads have a NULL mm, simply skip them, returning 0.
    if not mm:
        return TaskRss(0, 0, 0, 0)

    prog = task.prog_
    rss_stat = mm.rss_stat

    MM_FILEPAGES = prog.constant("MM_FILEPAGES").value_()
    MM_ANONPAGES = prog.constant("MM_ANONPAGES").value_()
    MM_SWAPENTS = prog.constant("MM_SWAPENTS").value_()

    # This counter was added in Linux kernel commit eca56ff906bd ("mm, shmem:
    # add internal shmem resident memory accounting") (in v4.5).
    try:
        MM_SHMEMPAGES = prog.constant("MM_SHMEMPAGES").value_()
    except LookupError:
        MM_SHMEMPAGES = -1

    if rss_stat.type_.kind == TypeKind.ARRAY:
        # Since Linux kernel commit f1a7941243c10 ("mm: convert mm's rss stats
        # into percpu_counter") (in v6.2), the "rss_stat" object is an array of
        # percpu counters. Simply sum them up!
        filerss = percpu_counter_sum(rss_stat[MM_FILEPAGES].address_of_())
        anonrss = percpu_counter_sum(rss_stat[MM_ANONPAGES].address_of_())
        swapents = percpu_counter_sum(rss_stat[MM_SWAPENTS].address_of_())
        shmemrss = percpu_counter_sum(rss_stat[MM_SHMEMPAGES].address_of_())
    else:
        # Prior to this, the "rss_stat" was a structure containing counters that
        # were cached on each task_struct and periodically updated into the
        # mm_struct. We start with the counter values from the mm_struct and
        # then sum up the cached copies from each thread.
        filerss = rss_stat.count[MM_FILEPAGES].counter.value_()
        anonrss = rss_stat.count[MM_ANONPAGES].counter.value_()
        swapents = rss_stat.count[MM_SWAPENTS].counter.value_()
        shmemrss = 0
        if MM_SHMEMPAGES >= 0:
            shmemrss = rss_stat.count[MM_SHMEMPAGES].counter.value_()

        for gtask in for_each_task_in_group(task, include_self=True):
            # Kernel configurations with a small NR_CPUS don't have the
            # per-thread cache.
            try:
                rss_stat = gtask.rss_stat
            except AttributeError:
                break
            filerss += rss_stat.count[MM_FILEPAGES].value_()
            anonrss += rss_stat.count[MM_ANONPAGES].value_()
            swapents += rss_stat.count[MM_SWAPENTS].value_()
            if MM_SHMEMPAGES >= 0:
                shmemrss += rss_stat.count[MM_SHMEMPAGES].value_()

    return TaskRss(filerss, anonrss, shmemrss, swapents)


def task_vsize(task: Object) -> int:
    """
    Return the virtual memory size of a task in bytes.

    :param task: ``struct task_struct *``
    """
    mm = task.mm.read_()
    if not mm:
        return 0
    return mm.total_vm.value_() * task.prog_["PAGE_SIZE"].value_()


@takes_program_or_default
def for_each_memory_block(prog: Program) -> Iterable[Object]:
    """
    Iterate over all memory hotplug blocks.

    :return: Iterator of ``struct memory_block *`` objects.
    """
    memory_block_type = prog.type("struct memory_block")
    for dev in bus_for_each_dev(prog["memory_subsys"].address_of_()):
        yield container_of(dev, memory_block_type, "dev")


def decode_memory_block_state(mem: Object) -> str:
    """
    Get a human-readable representation of the state of a memory hotplug block.

    >>> decode_memory_block_state(mem)
    'MEM_ONLINE'

    :param mem: ``struct memory_block *``
    """
    return decode_memory_block_state_value(mem.state)


_OLD_MEMORY_BLOCK_STATE = {
    # Added in Linux kernel commit 3947be1969a9 ("[PATCH] memory hotplug: sysfs
    # and add/remove functions") (in v2.6.15).
    (1 << 0): "MEM_ONLINE",
    (1 << 1): "MEM_GOING_OFFLINE",
    (1 << 2): "MEM_OFFLINE",
    # Added in Linux kernel commit 7b78d335ac15 ("memory hotplug: rearrange
    # memory hotplug notifier") (in v2.6.24).
    (1 << 3): "MEM_GOING_ONLINE",
    (1 << 4): "MEM_CANCEL_ONLINE",
    (1 << 5): "MEM_CANCEL_OFFLINE",
    # Added in Linux kernel commit c5f1e2d18909 ("mm/memory_hotplug: introduce
    # MEM_PREPARE_ONLINE/MEM_FINISH_OFFLINE notifiers") (in v6.5), removed in
    # 300709fbefd1 ("mm/memory_hotplug: Remove
    # MEM_PREPARE_ONLINE/MEM_FINISH_OFFLINE notifiers") (in v6.19).
    (1 << 6): "MEM_PREPARE_ONLINE",
    (1 << 7): "MEM_FINISH_OFFLINE",
}


@takes_program_or_default
def decode_memory_block_state_value(prog: Program, state: IntegerLike) -> str:
    """
    Get a human-readable representation of a memory hotplug block state value.

    >>> decode_memory_block_state_value(mem.state)
    'MEM_ONLINE'

    :param mem: ``enum memory_block_state`` or ``unsigned long``
    """
    # Since Linux kernel commit 1a4f70f6851a ("mm: convert memory block states
    # (MEM_*) macros to enum") (in v6.19), the states are in an enum. Before
    # that, we have to hard-code them.
    try:
        memory_block_state_type = prog.type("enum memory_block_state")
    except LookupError:
        return _OLD_MEMORY_BLOCK_STATE[operator.index(state)]
    else:
        if isinstance(state, Object):
            state = cast(memory_block_state_type, state)
        else:
            state = Object(prog, memory_block_state_type, state)
        return state.format_(type_name=False)


@takes_program_or_default
def memory_block_size_bytes(prog: Program) -> Object:
    """
    Return the size of a memory hotplug block in bytes.

    This is the unit that can be hot(un)plugged.

    :return: ``unsigned long``
    """
    return (
        cast("unsigned long", prog["sections_per_block"]) << prog["SECTION_SIZE_BITS"]
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import importlib
import pkgutil
import sys
import types
from typing import Any, Dict, Optional, Union

from drgn import Program
import drgn.cli
from drgn.commands import (
    Command,
    CommandNotFoundError,
    _parse_py_command,
    _print_py_command_exception,
    _write_command_error,
    argument,
)
from drgn.commands._crash._sys import _print_sys
from drgn.commands._crash.common import (
    CRASH_COMMAND_NAMESPACE,
    crash_command,
    crash_custom_command,
)
from drgn.commands.linux import linux_kernel_raw_command
from drgn.internal.repl import readline

# Import all submodules, recursively.
for _module_info in pkgutil.walk_packages(__path__, __name__ + "."):
    importlib.import_module(_module_info.name)


# These inherit from SystemExit to bypass things that attempt to handle most
# exceptions like the Python interactive console.
class _ExitToCrash(SystemExit):
    pass


class _ExitCrash(SystemExit):
    pass


def _crash_interactive_onerror(e: Exception) -> None:
    _write_command_error(sys.stderr, e, prefix="drgn: crash")


class _CrashCompleter:
    def __init__(self, prog: Program) -> None:
        self._prog = prog

    def complete(self, text: str, state: int) -> Optional[str]:
        # Only complete command names at the beginning of a line.
        begidx = readline.get_begidx()
        if begidx > 0:
            if readline.get_line_buffer()[:begidx].strip():
                return None

        if state == 0:
            self._matches = [
                name
                for name, _ in CRASH_COMMAND_NAMESPACE.enabled(self._prog)
                if name.startswith(text)
            ]

        if 0 <= state < len(self._matches):
            return self._matches[state]
        else:
            return None


@linux_kernel_raw_command(
    description="run a crash command",
    usage="**crash** [*command*]",
    long_description="""
    This provides a compatibility mode emulating the `crash utility
    <https://crash-utility.github.io/>`_.

    If *command* is given, run the given crash command (which may include
    arguments, redirections, pipes, etc.). Otherwise, enter an interactive
    prompt where crash commands can be called directly.

    Run ``%crash help`` or see :doc:`crash_compatibility` for the list of
    commands.
    """,
)
def _cmd_crash(
    prog: Program, name: str, args: str, *, globals: Dict[str, Any], **kwargs: Any
) -> Any:
    try:
        if args:
            return CRASH_COMMAND_NAMESPACE.run(prog, args, globals=globals)

        had_outer_repl = "outer_repl" in prog.config
        try:
            if not had_outer_repl:
                prog.config["outer_repl"] = "crash"
            elif prog.config["outer_repl"] == "crash":
                raise _ExitToCrash()

            with drgn.cli._setup_readline(
                drgn.cli._state_file("crash_history"), _CrashCompleter(prog).complete
            ):
                _print_sys(prog, context="panic")
                while True:
                    try:
                        line = input("%crash> ")
                        if not line or line.isspace():
                            continue
                    except EOFError:
                        break
                    except KeyboardInterrupt:
                        print("^C")
                        continue
                    try:
                        CRASH_COMMAND_NAMESPACE.run(
                            prog,
                            line,
                            globals=globals,
                            onerror=_crash_interactive_onerror,
                        )
                    except KeyboardInterrupt:
                        print()
                        continue
                    except _ExitToCrash:
                        continue
        finally:
            if not had_outer_repl:
                prog.config.pop("outer_repl", None)
    except _ExitCrash:
        pass


@crash_command(description="exit session")
@crash_command(name="q", description="exit session")
def _crash_cmd_exit(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    raise _ExitCrash()


@crash_custom_command(
    description="run drgn code or enter drgn interactive mode",
    usage="**drgn** [*code*]",
    long_description="""
    If *code* is given, execute the given drgn Python code, up to the first
    shell redirection or pipeline.

    Otherwise, enter drgn's interactive mode.
    """,
    parse=_parse_py_command,
)
def _crash_cmd_drgn(
    prog: Program,
    name: str,
    code: Union[types.CodeType, SyntaxError, None],
    *,
    globals: Dict[str, Any],
    **kwargs: Any,
) -> None:
    if code is None:
        if prog.config.get("outer_repl") == "drgn":
            raise _ExitCrash()
        else:
            drgn.cli.run_interactive(prog)
    elif isinstance(code, SyntaxError):
        _print_py_command_exception(code)
    else:
        try:
            exec(code, globals)
        except (Exception, KeyboardInterrupt) as e:
            _print_py_command_exception(e)


def _help_overview(prog: Program) -> None:
    command_names = [name for name, _ in CRASH_COMMAND_NAMESPACE.enabled(prog)]
    command_names.sort()
    columns = 5
    rows = (len(command_names) + columns - 1) // columns
    print("Commands:")
    for row in range(rows):
        print(
            "".join(
                [
                    command_names[i].ljust(
                        80 // columns if i + rows < len(command_names) else 0
                    )
                    for i in range(row, len(command_names), rows)
                ]
            )
        )
    print(
        """
Try "help <command>" for help with a specific command.
""",
        end="",
    )


def _print_help(name: str, command: Command[Any]) -> None:
    print(
        f"""\
NAME
  {name} - {command.description()}

SYNOPSIS
  {command.format_usage()}

DESCRIPTION
{command.format_help(indent="  ")}"""
    )


@crash_command(
    description="get help",
    long_description="""
    Print help about one or more commands.

    If no commands are given, this lists available commands and prints general
    help.
    """,
    arguments=(
        argument(
            "command",
            nargs="*",
            help="command to get help about, or ``all`` for all commands",
        ),
    ),
)
def _crash_cmd_help(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    if not args.command:
        return _help_overview(prog)

    first = True
    for name in args.command:
        if name == "all":
            for name, command in sorted(CRASH_COMMAND_NAMESPACE.enabled(prog)):
                if first:
                    first = False
                else:
                    print()
                _print_help(name, command)
        else:
            if first:
                first = False
            else:
                print()

            try:
                command = CRASH_COMMAND_NAMESPACE.lookup(prog, name)
            except CommandNotFoundError as e:
                print(f"No command named {e.name!r}")
            else:
                _print_help(name, command)



#!/usr/bin/env drgn
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""List the paths of all descendants of a cgroup v2"""

import argparse
from contextlib import contextmanager
import os
import sys
from collections import Counter

from drgn import cast
from drgn.helpers.common.type import enum_type_to_class
from drgn.helpers.linux import (
    cgroup_bpf_prog_for_each,
    cgroup_path,
    css_for_each_descendant_pre,
    fget,
    find_task,
)

# Since Linux kernel commit 6fc88c354f3a ("bpf: Migrate cgroup_bpf to internal
# cgroup_bpf_attach_type enum") (in v5.15), the attach type is the
# cgroup-specific enum cgroup_bpf_attach_type. Before that, it was the generic
# enum bpf_attach_type.
try:
    enum_cgroup_bpf_attach_type = prog.type("enum cgroup_bpf_attach_type")
except LookupError:
    CgroupBpfAttachType = enum_type_to_class(
        prog.type("enum bpf_attach_type"),
        "CgroupBpfAttachType",
        exclude=("__MAX_BPF_ATTACH_TYPE",),
    )
else:
    CgroupBpfAttachType = enum_type_to_class(
        enum_cgroup_bpf_attach_type,
        "CgroupBpfAttachType",
        exclude=("CGROUP_BPF_ATTACH_TYPE_INVALID", "MAX_CGROUP_BPF_ATTACH_TYPE",),
    )

CgroupSubsysId = enum_type_to_class(
    prog.type("enum cgroup_subsys_id"),
    "CgroupSubsysId",
    exclude=("CGROUP_SUBSYS_COUNT",),
)

@contextmanager
def open_dir(*args, **kwds):
    # Built-in open() context manager can't deal with directories.
    fd = os.open(*args, **kwds)
    try:
        yield fd
    finally:
        os.close(fd)


def print_cgroup_bpf_progs(cgrp):
    cgroup_printed = False
    for attach_type in CgroupBpfAttachType:
        attach_flags = cgrp.bpf.flags[attach_type.value].value_()
        for prog in cgroup_bpf_prog_for_each(cgrp, attach_type.value):
            prog_id = prog.aux.id.value_()
            prog_name = prog.aux.name.string_().decode()
            if not cgroup_printed:
                print(cgroup_path(cgrp).decode())
                cgroup_printed = True
            print(
                "    {:<8} {:<30} {:<15} {:<15}".format(
                    prog_id, attach_type.name, attach_flags, prog_name
                )
            )


def get_cgroup(path):
    task = find_task(prog, os.getpid())
    try:
        with open_dir(path, os.O_RDONLY) as fd:
            file_ = fget(task, fd)
            kn = cast("struct kernfs_node *", file_.f_path.dentry.d_inode.i_private)
            return cast("struct cgroup *", kn.priv)
    except FileNotFoundError as e:
        raise argparse.ArgumentTypeError(e)


def cmd_tree(cgroup):
    css = cgroup.self.address_of_()

    for pos in css_for_each_descendant_pre(css):
        if not pos.flags & prog["CSS_ONLINE"]:
            continue
        print(cgroup_path(pos.cgroup).decode())


def cmd_bpf(cgroup):
    css = cgroup.self.address_of_()

    for pos in css_for_each_descendant_pre(css):
        if not pos.flags & prog["CSS_ONLINE"]:
            continue
        print_cgroup_bpf_progs(pos.cgroup)


def cmd_stat(cgroup):
    stat = Counter()
    stat_dying = Counter()

    for ssid in CgroupSubsysId:
        css = cgroup.subsys[ssid.value]
        # XXX if subsys of offlined or cgroup rmdir'd under our hands we won't see its subtree
        if not css:
            continue
        for pos in css_for_each_descendant_pre(css):
            stat[ssid] +=1
            if not pos.flags & prog["CSS_ONLINE"]:
                stat_dying[ssid] += 1

    for ssid in CgroupSubsysId:
        if stat[ssid.value] == 0:
            continue
        print("nr_{:<30} {:>4}".format(
            ssid.name,
            stat[ssid.value]
            )
        )
    for ssid in CgroupSubsysId:
        if stat_dying[ssid.value] == 0:
            continue
        print("nr_dying_{:<24} {:>4}".format(
            ssid.name,
            stat_dying[ssid.value]
            )
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("command", choices=["tree", "bpf", "stat"])
    parser.add_argument("cgroups", help="Cgroups", nargs="*", type=get_cgroup)
    args = parser.parse_args()

    if len(args.cgroups) == 0:
        args.cgroups.append(prog["cgrp_dfl_root"].cgrp)

    for cg in args.cgroups:
        if len(args.cgroups) > 1:
            print(cg.kn.name.string_())
        locals()["cmd_" + args.command](cg)




# Copyright (c) Meta Platforms, Inc. and affiliates.
# Copyright (c) 2025, Kylin Software, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Memory management-related crash commands (other than kmem)."""

import argparse
import sys
from typing import Any, List, Sequence

from drgn import Architecture, FaultError, Object, Platform, Program
from drgn.commands import (
    DrgnCodeBlockContext,
    argument,
    argument_group,
    drgn_argument,
    mutually_exclusive_group,
)
from drgn.commands._crash.common import (
    CrashDrgnCodeBuilder,
    _crash_foreach_subcommand,
    _TaskSelector,
    crash_command,
    parse_cpuspec,
    print_task_header,
)
from drgn.helpers.common.format import CellFormat, escape_ascii_string, print_table
from drgn.helpers.linux.mm import (
    follow_phys,
    for_each_vma,
    phys_to_virt,
    task_rss,
    vma_name,
)
from drgn.helpers.linux.percpu import per_cpu_ptr
from drgn.helpers.linux.swap import (
    for_each_swap_info,
    swap_file_path,
    swap_is_file,
    swap_usage_in_pages,
)


@crash_command(
    description="bytes to page",
    long_description="Convert byte numbers (usually physical addresses) to page numbers.",
    arguments=(
        argument(
            "address",
            type="hexadecimal",
            nargs="+",
            help="hexadecimal byte number/physical address",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_btop(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import("drgn.helpers.linux.mm", "PHYS_PFN")

        if len(args.address) == 1:
            code.append(f"phys_addr = {hex(args.address[0])}\n")
        else:
            code.append("for phys_addr in ")
            for i, address in enumerate(args.address):
                if i > 0:
                    code.append(", ")
                code.append(hex(address))
            code.append(":\n    ")

        code.append("pfn = PHYS_PFN(phys_addr)\n")
        code.print()
        return

    PAGE_SHIFT = prog["PAGE_SHIFT"].value_()
    for address in args.address:
        print(f"{address:x}: {address >> PAGE_SHIFT:x}")


@crash_command(
    description="page to bytes",
    long_description="Convert page frame numbers to byte numbers (physical addresses).",
    arguments=(
        argument(
            "pfn", type="decimal_or_hexadecimal", nargs="+", help="page frame number"
        ),
        drgn_argument,
    ),
)
def _crash_cmd_ptob(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import("drgn.helpers.linux.mm", "PFN_PHYS")

        if len(args.pfn) == 1:
            code.append(f"pfn = {hex(args.pfn[0])}\n")
        else:
            code.append("for pfn in ")
            for i, pfn in enumerate(args.pfn):
                if i > 0:
                    code.append(", ")
                code.append(hex(pfn))
            code.append(":\n    ")

        code.append("phys_addr = PFN_PHYS(pfn)\n")
        code.print()
        return

    PAGE_SHIFT = prog["PAGE_SHIFT"].value_()
    for pfn in args.pfn:
        print(f"{pfn:x}: {pfn << PAGE_SHIFT:x}")


@crash_command(
    description="physical or per-CPU to virtual",
    long_description="""This command translates a hexadecimal physical address into a
                        kernel virtual address. Alternatively, a hexadecimal per-cpu
                        offset and cpu specifier will be translated into kernel virtual
                        addresses for each cpu specified.""",
    arguments=(
        argument(
            "address",
            metavar="address|offset:cpuspec",
            nargs="+",
            help="hexadecimal physical address or hexadecimal per-CPU offset and CPU specifier",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_ptov(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    if args.drgn:
        # Create a single builder for all addresses
        builder = CrashDrgnCodeBuilder(prog)
        physical_addresses = []
        per_cpu_offsets = []

        for address in args.address:
            if ":" in address:
                # Add imports only once
                builder.add_from_import("drgn", "Object")
                builder.add_from_import("drgn.helpers.linux.percpu", "per_cpu_ptr")
                builder.add_from_import(
                    "drgn.helpers.linux.cpumask", "for_each_possible_cpu"
                )
                # Parse the cpuspec in the actual command code
                offset_str, cpu_spec = address.split(":", 1)
                offset = int(offset_str, 16)
                per_cpu_offsets.append((offset, parse_cpuspec(cpu_spec)))
            else:
                # Add imports only once
                builder.add_from_import("drgn.helpers.linux.mm", "phys_to_virt")
                physical_addresses.append(int(address, 16))

        # Generate code for physical addresses
        if physical_addresses:
            builder.append("addresses = [")
            builder.append(", ".join(f"0x{addr:x}" for addr in physical_addresses))
            builder.append("]\n")
            builder.append("for address in addresses:\n")
            builder.append("    virt = phys_to_virt(address)\n")

        # Generate code for per-CPU offsets
        for offset, cpuspec in per_cpu_offsets:
            builder.append(f"\noffset = {offset:#x}\n")
            with builder.begin_cpuspec_loop(cpuspec):
                builder.append(
                    'virt = per_cpu_ptr(Object(prog, "void *", offset), cpu)\n'
                )

        # Print the generated code once at the end
        builder.print()
        return

    # Handle direct execution without --drgn
    for i, address in enumerate(args.address):
        if i > 0:
            print()  # Add a blank line between outputs for multiple addresses

        if ":" in address:
            # Handle per-CPU offset case
            offset_str, cpu_spec = address.split(":", 1)
            offset = int(offset_str, 16)

            # Parse CPU specifier using parse_cpuspec
            cpus = parse_cpuspec(cpu_spec)

            # Print offset information
            print(f"PER-CPU OFFSET: {offset:x}")  # Directly print offset information

            # Prepare data for print_table()
            rows = [("  CPU", "  VIRTUAL")]  # Add CPU and VIRTUAL header
            ptr = Object(prog, "void *", offset)  # Changed type to "void *"
            for cpu in cpus.cpus(prog):
                virt = per_cpu_ptr(ptr, cpu)
                rows.append((f"  [{cpu}]", f"{virt.value_():016x}"))

            # Print the table
            print_table(rows)
        else:
            # Handle physical address case
            phys = int(address, 16)
            virt = phys_to_virt(prog, phys)
            virt_int = virt.value_()

            # Prepare data for print_table()
            rows = [("VIRTUAL", "PHYSICAL"), (f"{virt_int:016x}", f"{phys:x}")]

            # Print the table
            print_table(rows)


# This is not an exhaustive list.
def _overlapping_address_spaces(prog: Program) -> bool:
    platform: Platform = prog.platform  # type: ignore[assignment]  # platform can't be None.
    return platform.arch == Architecture.S390X


@_crash_foreach_subcommand(
    arguments=(
        mutually_exclusive_group(
            argument(
                "-u",
                dest="user",
                action="store_true",
            ),
            argument(
                "-k",
                dest="kernel",
                action="store_true",
            ),
        ),
        argument(
            "addresses",
            metavar="address",
            type="hexadecimal",
            nargs="+",
        ),
        drgn_argument,
    ),
)
def _crash_foreach_vtop(
    task_selector: _TaskSelector, args: argparse.Namespace, *, task_headers: bool = True
) -> None:
    prog = task_selector.prog

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import("drgn", "FaultError")
        code.add_from_import("drgn.helpers.linux.mm", "follow_phys")

        if len(args.addresses) == 1:
            code.append(f"virtual_address = {hex(args.addresses[0])}\n\n")
        else:
            code.append(
                "virtual_addresses = ("
                + ", ".join([hex(address) for address in args.addresses])
                + ")\n\n"
            )

        def virtual_address_loop() -> DrgnCodeBlockContext:
            if len(args.addresses) == 1:
                return code.begin_block("")
            else:
                return code.begin_block("for virtual_address in virtual_addresses:\n")

        if not args.user:
            if not args.kernel:
                code.append("# As kernel address.\n")
            code.append('init_mm = prog["init_mm"].address_of_()\n')
            with virtual_address_loop():
                code.append(
                    """\
try:
    physical_address = follow_phys(init_mm, virtual_address)
except FaultError:
    pass
"""
                )

        if not args.kernel:
            if not args.user:
                code.append("\n# As user address.\n")
            with task_selector.begin_task_loop(code):
                if task_headers:
                    code.append_task_header()

                with code.begin_block(
                    """\
mm = task.mm.read_()
if mm:
"""
                ), virtual_address_loop():
                    code.append(
                        """\
try:
    physical_address = follow_phys(mm, virtual_address)
except FaultError:
    pass
"""
                    )
        return code.print()

    overlapping = _overlapping_address_spaces(prog)
    init_mm = prog["init_mm"].address_of_()

    first_task = True
    for task in task_selector.tasks():
        if first_task:
            first_task = False
        else:
            print()
        if task_headers:
            print_task_header(task)

        mm = task.mm.read_()
        if not args.kernel and not args.user and mm and not overlapping:
            task_size = mm.task_size.value_()

        first_address = True
        for address in args.addresses:
            if first_address:
                first_address = False
            else:
                print()

            kernel_phys_addr = None
            user_phys_addr = None
            if args.kernel or (
                not args.user and (not mm or overlapping or address >= task_size)
            ):
                try:
                    kernel_phys_addr = follow_phys(init_mm, address).value_()
                except FaultError:
                    pass
            if args.user or (
                not args.kernel and mm and (overlapping or address < task_size)
            ):
                try:
                    user_phys_addr = follow_phys(mm, address).value_()
                except FaultError:
                    pass

            if kernel_phys_addr is not None and user_phys_addr is not None:
                print(
                    "error: ambiguous address: %x: -u or -k is required",
                    address,
                    file=sys.stderr,
                )
                continue
            elif kernel_phys_addr is None and user_phys_addr is None:
                phys_addr_cell: Any = "(not accessible)"
            else:
                phys_addr_cell = CellFormat(
                    kernel_phys_addr if user_phys_addr is None else user_phys_addr, "<x"
                )
            print_table(
                (
                    ("VIRTUAL", "PHYSICAL"),
                    (CellFormat(address, "<x"), phys_addr_cell),
                )
            )


@crash_command(
    description="virtual to physical address",
    arguments=(
        argument_group(
            mutually_exclusive_group(
                argument(
                    "-u",
                    dest="user",
                    action="store_true",
                    help="treat addresses as user space virtual addresses",
                ),
                argument(
                    "-k",
                    dest="kernel",
                    action="store_true",
                    help="treat addresses as kernel space virtual addresses",
                ),
            ),
            title="address space",
            description="""
            In some kernel configurations, whether an address belongs to kernel
            space or user space can be determined from only its value. In those
            cases, this command will automatically determine whether the
            address belongs to kernel or user space.

            In other configurations (including s390x, SPARC, and 4G:4G), this
            is not possible, and one of these options is required.
            """,
        ),
        argument(
            "-c",
            dest="task",
            type="pid_or_task",
            help="""
            PID or hexadecimal task_struct pointer of task whose address space
            should be used for translating virtual addresses
            """,
        ),
        argument(
            "addresses",
            metavar="address",
            type="hexadecimal",
            nargs="+",
            help="hexadecimal virtual address",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_vtop(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    return _crash_foreach_vtop(
        _TaskSelector(prog, [args.task]), args, task_headers=False
    )


@crash_command(
    description="list swap devices",
    arguments=(drgn_argument,),
)
def _crash_cmd_swap(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import(
            "drgn.helpers.linux.swap",
            "for_each_swap_info",
            "swap_file_path",
            "swap_is_file",
            "swap_usage_in_pages",
        )
        code.append(
            """\
for si in for_each_swap_info():
    is_file = swap_is_file(si)
    pages = si.pages
    used_pages = swap_usage_in_pages(si)
    priority = si.prio
    path = swap_file_path(si)
"""
        )
        code.print()
        return

    rows: List[Sequence[Any]] = [
        (
            "SWAP_INFO_STRUCT",
            CellFormat("TYPE", "^"),
            CellFormat("SIZE", "^"),
            CellFormat("USED", "^"),
            CellFormat("PCT", ">"),
            CellFormat("PRI", ">"),
            "FILENAME",
        ),
    ]
    for si in for_each_swap_info(prog):
        size = (si.pages * prog["PAGE_SIZE"]).value_()
        used = swap_usage_in_pages(si) * prog["PAGE_SIZE"].value_()
        rows.append(
            (
                CellFormat(si.value_(), "x"),
                CellFormat(
                    "FILE" if swap_is_file(si) else "PARTITION",
                    "^",
                ),
                CellFormat(f"{size // 1024}k", "^"),
                CellFormat(f"{used // 1024}k", "^"),
                CellFormat(used / size, ".0%"),
                si.prio.value_(),
                escape_ascii_string(swap_file_path(si)),
            ),
        )
    print_table(rows)


@_crash_foreach_subcommand(
    arguments=(drgn_argument,),
)
def _crash_foreach_vm(task_selector: _TaskSelector, args: argparse.Namespace) -> None:
    prog = task_selector.prog

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        with task_selector.begin_task_loop(code):
            code.append_task_header()
            code.add_from_import(
                "drgn.helpers.linux.mm", "for_each_vma", "task_rss", "vma_name"
            )
            code.append(
                """\

mm = task.mm.read_()
if mm:
    pgd = mm.pgd
    rss = task_rss(task)
    total_vm = mm.total_vm

    for vma in for_each_vma(mm):
        start = vma.vm_start
        end = vma.vm_end
        flags = vma.vm_flags
        file = vma_name(vma)
"""
            )
        return code.print()

    first = True
    for task in task_selector.tasks():
        if first:
            first = False
        else:
            print()
        print_task_header(task)

        mm = task.mm.read_()
        if mm:
            pgd_value = mm.pgd.value_()
            rss_total = task_rss(task).total
            total_vm = mm.total_vm.value_()
        else:
            pgd_value = rss_total = total_vm = 0
        page_size = prog["PAGE_SIZE"].value_()
        print_table(
            (
                (
                    CellFormat("MM", "^"),
                    CellFormat("PGD", "^"),
                    CellFormat("RSS", "^"),
                    CellFormat("TOTAL_VM", "^"),
                ),
                (
                    CellFormat(mm.value_(), "^x"),
                    CellFormat(pgd_value, "^x"),
                    CellFormat(f"{rss_total * page_size // 1024}k", "^"),
                    CellFormat(f"{total_vm * page_size // 1024}k", "^"),
                ),
            )
        )
        if not mm:
            return

        rows: List[Sequence[Any]] = [
            (
                CellFormat("VMA", "^"),
                CellFormat("START", "^"),
                CellFormat("END", "^"),
                CellFormat("FLAGS", "<"),
                CellFormat("FILE", "<"),
            )
        ]
        for vma in for_each_vma(mm):
            rows.append(
                (
                    CellFormat(vma.value_(), "^x"),
                    CellFormat(vma.vm_start.value_(), "^x"),
                    CellFormat(vma.vm_end.value_(), "^x"),
                    CellFormat(vma.vm_flags.value_(), "<x"),
                    escape_ascii_string(vma_name(vma), escape_backslash=True),
                )
            )
        print_table(rows)


@crash_command(
    description="virtual memory",
    long_description="""This command displays basic virtual memory information of a context,
consisting of a pointer to its mm_struct and page directory, its RSS and
total virtual memory size; and a list of pointers to each vm_area_struct,
its starting and ending address, vm_flags value, and file pathname. If no
arguments are entered, the current context is used.
""",
    arguments=(
        argument(
            "tasks",
            metavar="pid|task",
            nargs="*",
            type="pid_or_task",
            help="one or more process PIDs or hexadecimal task_struct pointers",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_vm(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    if not args.tasks:
        args.tasks.append(None)
    return _crash_foreach_vm(_TaskSelector(prog, args.tasks), args)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Linux Kernel
------------

The ``drgn.helpers.linux`` package contains several modules for working with
data structures and subsystems in the Linux kernel. The helpers are available
from the individual modules in which they are defined and from this top-level
package. E.g., the following are both valid:

>>> from drgn.helpers.linux.list import list_for_each_entry
>>> from drgn.helpers.linux import list_for_each_entry

Iterator macros (``for_each_foo``) are a common idiom in the Linux kernel. The
equivalent drgn helpers are implemented as Python :ref:`generators
<python:tut-generators>`. For example, the following code in C:

.. code-block:: c

    list_for_each(pos, head)
            do_something_with(pos);

Translates to the following code in Python:

.. code-block:: python3

    for pos in list_for_each(head):
        do_something_with(pos)
"""

import importlib
import pkgutil
from typing import List

__all__: List[str] = []
for _module_info in pkgutil.iter_modules(__path__, prefix=__name__ + "."):
    _submodule = importlib.import_module(_module_info.name)
    _submodule_all = getattr(_submodule, "__all__", ())
    __all__.extend(_submodule_all)
    for _name in _submodule_all:
        globals()[_name] = getattr(_submodule, _name)
